#! /usr/bin/env python
"""This module implements the IMS CP 1.2 specification defined by IMS GLC
"""

import itertools
import random
import re
import zipfile

from . import imsmdv1p2p1 as imsmd
from . import rfc2396 as uri
from . import vfs
from .pep8 import MigratedClass, old_function, old_method
from .py2 import py2, to_text
from .qtiv2 import xml as qtixml2
from .qtiv2 import metadata as qtimd
from .xml import structures as xml
from .xml import namespace as xmlns
from .xml import xsdatatypes as xsi


# :	String constant for the main namespace
IMSCP_NAMESPACE = "http://www.imsglobal.org/xsd/imscp_v1p1"
# :	String constant for the official schema location
IMSCP_SCHEMALOCATION = "http://www.imsglobal.org/xsd/imscp_v1p1.xsd"
# :	String constant for the 1.2 extension elements' namespace
IMSCPX_NAMESPACE = "http://www.imsglobal.org/xsd/imscp_extensionv1p2"

IGNOREFILES_RE = "\\..*"


class CPException(Exception):
    pass


class CPFilePathError(Exception):
    pass


class CPFileTypeError(Exception):
    pass


class CPManifestError(CPException):
    pass


class CPProtocolError(CPException):
    pass


class CPValidationError(CPException):
    pass


class CPPackageBeenThereError(CPException):
    pass


class CPZIPBeenThereError(Exception):
    pass


class CPZIPDirectorySizeError(CPException):
    pass


class CPZIPDuplicateFileError(CPException):
    pass


class CPZIPFilenameError(CPException):
    pass


class CPElement(xmlns.XMLNSElement):

    """Base class for all elements defined by the Content Packaging
    specification."""
    pass


@old_function('PathInPath')
def path_in_path(child_path, parent_path):
    """Utility function that returns child_path expressed relative to parent_path

    This function processes file system paths, not the path components of URI.

    Both paths are normalized to remove any redundant navigational segments
    before any processing, the resulting path will not contain these either.

    If child_path is not contained in parent_path then None is returned.

    If child_path and parent_path are equal an empty string is returned."""
    rel_path = []
    child_path = child_path.normpath()
    parent_path = parent_path.normpath().normcase()
    while child_path.normcase() != parent_path:
        child_path, tail = child_path.split()
        if not child_path or not tail:
            # We've gone as far as we can, fail!
            return None
        rel_path[0:0] = [tail]
    if rel_path:
        return parent_path.__class__(*rel_path)
    else:
        return ''


class Schema(CPElement):

    """Represents the schema element."""
    XMLNAME = (IMSCP_NAMESPACE, 'schema')


class SchemaVersion(CPElement):

    """Represents the schemaversion element."""
    XMLNAME = (IMSCP_NAMESPACE, 'schemaversion')


class Metadata(CPElement):

    """Represents the Metadata element."""
    XMLNAME = (IMSCP_NAMESPACE, 'metadata')
    XMLCONTENT = xml.ElementContent

    #: the default class to represent the schema element
    SchemaClass = Schema
    # : the default class to represent the schemaVersion element
    SchemaVersionClass = SchemaVersion

    def __init__(self, parent):
        CPElement.__init__(self, parent)
        #: the optional schema element
        self.Schema = None
        #: the optional schemaversion element
        self.SchemaVersion = None

    def get_children(self):
        if self.Schema:
            yield self.Schema
        if self.SchemaVersion:
            yield self.SchemaVersion
        for child in CPElement.get_children(self):
            yield child


class Organization(CPElement):

    """Represents the organization element."""
    XMLNAME = (IMSCP_NAMESPACE, 'organization')


class Organizations(CPElement):

    """Represents the organizations element."""
    XMLNAME = (IMSCP_NAMESPACE, 'organizations')
    XMLCONTENT = xml.ElementContent

    #: the default class to represent the organization element
    OrganizationClass = Organization

    def __init__(self, parent):
        CPElement.__init__(self, parent)
        #: a list of organization elements
        self.Organization = []

    def get_children(self):
        return itertools.chain(self.Organization, CPElement.get_children(self))


class File(CPElement):

    """Represents the file element."""
    XMLNAME = (IMSCP_NAMESPACE, 'file')
    XMLATTR_href = ('href', uri.URI.from_octets, str)

    def __init__(self, parent):
        CPElement.__init__(self, parent)
        #: the href used to locate the file object
        self.href = None

    @old_method('PackagePath')
    def package_path(self, cp):
        """Returns the normalized file path relative to the root of the content
        package, *cp*.

        If the href does not point to a local file then None is returned.
        Otherwise, this function calculates an absolute path to the file and
        then calls the content package's :py:meth:`ContentPackage.package_path`
        method."""
        url = self.resolve_uri(self.href)
        if not isinstance(url, uri.FileURL):
            return None
        return cp.package_path(url.get_virtual_file_path())


class Dependency(CPElement):

    """Represents the dependency element."""
    XMLNAME = (IMSCP_NAMESPACE, 'dependency')
    XMLATTR_identifierref = 'identifierref'

    def __init__(self, parent):
        CPElement.__init__(self, parent)
        #: the identifier of the resource in this dependency
        self.identifierref = None


class Resource(CPElement):

    """Represents the resource element."""
    XMLNAME = (IMSCP_NAMESPACE, 'resource')
    ID = (xmlns.NO_NAMESPACE, "identifier")
    XMLATTR_href = ('href', uri.URI.from_octets, str)
    XMLATTR_type = 'type'
    XMLCONTENT = xml.ElementContent

    #: the default class to represent the metadata element
    MetadataClass = Metadata
    #: the default class to represent the file element
    FileClass = File
    #: the default class to represent the dependency element
    DependencyClass = Dependency

    def __init__(self, parent):
        CPElement.__init__(self, parent)
        #: the type of the resource
        self.type = None
        #: the href pointing at the resource's entry point
        self.href = None
        #: the resource's optional metadata element
        self.Metadata = None
        #: a list of file elements associated with the resource
        self.File = []
        #: a list of dependencies of this resource
        self.Dependency = []

    @old_method('GetEntryPoint')
    def get_entry_point(self):
        """Returns the :py:class:`File` object that is identified as the entry point.

        If there is no entry point, or no :py:class:`File` object with a
        matching href, then None is returned."""
        href = self.href
        if href:
            href = self.resolve_uri(href)
            for f in self.File:
                f_href = f.href
                if f_href:
                    f_href = f.resolve_uri(f_href)
                    if href.match(f_href):
                        return f
        return None

    @old_method('SetEntryPoint')
    def set_entry_point(self, f):
        """Set's the :py:class:`File` object that is identified as the
        resource's entry point.

        The File must already exist and be associated with the resource."""
        # We resolve and recalculate just in case xml:base lurks on this file
        href = self.relative_uri(f.resolve_uri(f.href))
        self.href = href

    def get_children(self):
        if self.Metadata:
            yield self.Metadata
        for child in itertools.chain(
                self.File,
                self.Dependency,
                CPElement.get_children(self)):
            yield child

    @old_method('DeleteFile')
    def delete_file(self, f):
        index = self.File.index(f)
        f.detach_from_doc()
        f.parent = None
        del self.File[index]

    @old_method('DeleteDependency')
    def delete_dependency(self, d):
        index = self.Dependency.index(d)
        d.detach_from_doc()
        d.parent = None
        del self.Dependency[index]


class Resources(CPElement):

    """Represents the resources element."""
    XMLNAME = (IMSCP_NAMESPACE, 'resources')
    XMLCONTENT = xml.ElementContent

    #: the default class to represent the resource element
    ResourceClass = Resource

    def __init__(self, parent):
        CPElement.__init__(self, parent)
        #: the list of resources in the manifest
        self.Resource = []

    def get_children(self):
        return itertools.chain(self.Resource, CPElement.get_children(self))


class Manifest(CPElement):

    """Represents the manifest element, the root element of the imsmanifest
    file."""
    ID = (xmlns.NO_NAMESPACE, "identifier")
    XMLNAME = (IMSCP_NAMESPACE, 'manifest')
    XMLCONTENT = xml.ElementContent

    #: the default class to represent the metadata element
    MetadataClass = Metadata
    #: the default class to represent the organizations element
    OrganizationsClass = Organizations
    #: the default class to represent the resources element
    ResourcesClass = Resources
    #: the default class to represent child manifest elements
    ManifestClass = None

    def __init__(self, parent):
        CPElement.__init__(self, parent)
        #: the manifest's metadata element
        self.Metadata = None
        #: the organizations element
        self.Organizations = self.OrganizationsClass(self)
        #: the resources element
        self.Resources = self.ResourcesClass(self)
        #: a list of child manifest elements
        self.Manifest = []

    def get_children(self):
        if self.Metadata:
            yield self.Metadata
        if self.Organizations:
            yield self.Organizations
        if self.Resources:
            yield self.Resources
        for child in itertools.chain(
                self.Manifest,
                CPElement.get_children(self)):
            yield child

Manifest.ManifestClass = Manifest


class ManifestDocument(xmlns.XMLNSDocument):

    """Represents the imsmanifest.xml file itself.

    Buildong on :py:class:`pyslet.xml.namespace.XMLNSDocument` this class is
    used for parsing and writing manifest files.

    The constructor defines three additional prefixes using
    :py:meth:`~pyslet.xml.namespace.XMLNSDocument.make_prefix`, mapping xsi
    onto XML schema, imsmd onto the IMS LRM namespace and imsqti onto the IMS
    QTI 2.1 namespace.  It also adds a schemaLocation attribute.  The elements
    defined by the :py:mod:`pyslet.imsmdv1p2p1` and :py:mod:`pyslet.imsqtiv2p1`
    modules are added to the :py:attr:`classMap` to ensure that metadata from
    those schemas are bound to the special classes defined there."""

    classMap = {}

    def __init__(self, **args):
        xmlns.XMLNSDocument.__init__(self, **args)
        #: the default namespace is set to :py:const:`IMSCP_NAMESPACE`
        self.defaultNS = IMSCP_NAMESPACE
        self.make_prefix(xsi.XMLSCHEMA_NAMESPACE, 'xsi')
        self.make_prefix(imsmd.IMSLRM_NAMESPACE, 'imsmd')
        self.make_prefix(qtixml2.core.IMSQTI_NAMESPACE, 'imsqti')
        schema_location = [
            IMSCP_NAMESPACE, IMSCP_SCHEMALOCATION,
            imsmd.IMSLRM_NAMESPACE, imsmd.IMSLRM_SCHEMALOCATION,
            qtixml2.core.IMSQTI_NAMESPACE,
            qtixml2.core.IMSQTI_SCHEMALOCATION]
        if isinstance(self.root, CPElement):
            self.root.set_attribute(
                (xsi.XMLSCHEMA_NAMESPACE, 'schemaLocation'),
                ' '.join(schema_location))

    def get_element_class(self, name):
        """Overrides
        :py:meth:`pyslet.xml.namespace.XMLNSDocument.get_element_class` to look
        up name.

        The class contains a mapping from (namespace,element name) pairs to
        class objects representing the elements.  Any element not in the class
        map returns :py:meth:`~pyslet.xml.namespace.XMLNSElement` instead."""
        eclass = ManifestDocument.classMap.get(
            name, ManifestDocument.classMap.get((name[0], None),
                                                xmlns.XMLNSElement))
        return eclass

xmlns.map_class_elements(ManifestDocument.classMap, globals())
xmlns.map_class_elements(ManifestDocument.classMap, imsmd)
xmlns.map_class_elements(ManifestDocument.classMap, qtimd)
# Add other supported metadata schemas in here


class ContentPackage(MigratedClass):

    """Represents a content package.

    When constructed with no arguments a new package is created.  A temporary
    folder to hold the contents of the package is created and will not be
    cleaned up until the :py:meth:`Close` method is called.

    Alternatively, you can pass an operating system or virtual file path to a
    content package directory, to an imsmanifest.xml file or to a Package
    Interchange Format file.  In the latter case, the file is unzipped into a
    temporary folder to facilitate manipulation of the package contents.

    A new manifest file is created and written to the file system when creating
    a new package, or if it is missing from an existing package or
    directory."""

    #: the default class for representing the Manifest file
    ManifestDocumentClass = ManifestDocument

    def __init__(self, dpath=None):
        self.tempDir = False
        error_flag = True
        try:
            if dpath is None:
                #: the :py:class:`~pyslet.vfs.VirtualFilePath` to the package's
                # directory
                self.dPath = vfs.defaultFS.mkdtemp('.d', 'imscpv1p2-')
                self.tempDir = True
                self.packageName = 'imscp'
            else:
                if not isinstance(dpath, vfs.VirtualFilePath):
                    dpath = vfs.defaultFS(dpath)
                self.dPath = dpath.abspath()
                head, tail = self.dPath.split()
                self.packageName = tail
                if self.dPath.isdir():
                    # existing directory
                    pass
                elif self.dPath.exists():
                    # is this a zip archive?
                    f = self.dPath.open("rb")
                    try:
                        if zipfile.is_zipfile(f):
                            name, ext = tail.splitext()
                            if ext.lower() == ".zip":
                                self.packageName = name
                            self.expand_zip(f)
                        else:
                            # anything else must be a manifest file
                            self.dPath = head
                            mpath = tail
                            head, tail = self.dPath.split()
                            if str(mpath.normcase()) != 'imsmanifest.xml':
                                raise CPManifestError(
                                    "%s must be named imsmanifest.xml" %
                                    str(mpath))
                            self.packageName = str(tail)
                    finally:
                        f.close()
                else:
                    self.dPath.mkdir()
            if not isinstance(self.dPath, vfs.VirtualFilePath):
                import traceback
                traceback.print_stack()
            mpath = self.dPath.join('imsmanifest.xml')
            if mpath.exists():
                self.manifest = self.ManifestDocumentClass(
                    base_uri=str(uri.URI.from_virtual_path(mpath)))
                """The :py:class:`ManifestDocument` object representing the
                imsmanifest.xml file.

                The file is read (or created) on construction."""
                self.manifest.read()
                if not isinstance(self.manifest.root, Manifest):
                    raise CPManifestError("%s not a manifest file, found"
                                          " %s::%s " %
                                          (mpath, self.manifest.root.ns,
                                           self.manifest.root.xmlname))
            else:
                self.manifest = self.ManifestDocumentClass(
                    root=Manifest,
                    base_uri=str(uri.URI.from_virtual_path(mpath)))
                self.manifest.root.set_id(
                    self.manifest.get_unique_id('manifest'))
                md = self.manifest.root.add_child(
                    self.manifest.root.MetadataClass)
                md.add_child(md.SchemaClass).set_value("IMS Content")
                md.add_child(md.SchemaVersionClass).set_value("1.2")
                self.manifest.create()
            self.set_ignore_files(IGNOREFILES_RE)
            self.fileTable = {}
            """The fileTable is a dictionary that maps package relative file
            paths to the :py:class:`File` objects that represent them in the
            manifest.

            It is possible for a file to be referenced multiple times (although
            dependencies were designed to take care of most cases it is still
            possible for two resources to share a physical file, or even for a
            resource to contain multiple references to the same file.)
            Therefore, the dictionary values are lists of :py:class:`File`
            objects.

            If a file path maps to an empty list then a file exists in the
            package which is not referenced by any resource.  In some packages
            it is common for auxiliary files such as supporting schemas to be
            included in packages without a corresponding :py:class:`File`
            object so an empty list does not indicate that the file can be
            removed safely.  These files are still included when packaging the
            content package for interchange.

            Finally, if a file referred to by a :py:class:`File` object in the
            manifest is missing an entry is still created in the fileTable. You
            can walk the keys of the fileTable testing if each file exists to
            determine if some expected files are missing from the package.

            The keys in fileTable are VirtualFilePath instances.  To convert a
            string to an appropriate instance use the :py:meth:`FilePath`
            method."""
            self.rebuild_file_table()
            error_flag = False
        finally:
            if error_flag:
                self.Close()

    @old_method('FilePath')
    def make_file_path(self, *path):
        """Converts a string into a :py:class:`pyslet.vfs.VirtualFilePath`
        instance suitable for using as a key into the :py:attr:`fileTable`. The
        conversion is done using the file system of the content package's
        directory, :py:attr:`dPath`."""
        return self.dPath.__class__(*path)

    @old_method('SetIgnoreFiles')
    def set_ignore_files(self, ignore_files):
        """Sets the regular expression used to determine if a file should be ignored.

        Some operating systems and utilities create hidden files or other
        spurious data inside the content package directory.  For example,
        's OS X creates .DS_Store files and the svn source control utility
        creates .svn directories.  The files shouldn't generally be included in
        exported packages as they may confuse the recipient (who may be using a
        system on which these files and directories are not hidden) and be
        deemed to violate the specification, not to mention adding
        unnecessarily to the size of the package and perhaps even leaking
        information unintentionally.

        To help avoid this type of problem the class uses a regular expression
        to determine if a file should be considered part of the package.  When
        listing directories, the names of the files found are compared against
        this regular expression and are ignored if they match.

        By default, the pattern is set to match all directories and files with
        names beginning '.' so you will not normally need to call this
        method."""
        self.ignore_files = re.compile(ignore_files)

    @old_method('IgnoreFile')
    def ignore_file(self, f):
        """Compares a file or directory name against the pattern set by
        :py:meth:`set_ignore_files`.

        f is a unicode string."""
        match = self.ignore_files.match(f)
        if match:
            return len(f) == match.end()
        else:
            return False

    @old_method('IgnoreFilePath')
    def ignore_file_path(self, fpath):
        """Compares a file path against the pattern set by :py:meth:`set_ignore_files`

        The path is normalised before comparison and any segments consisting of
        the string '..' are skipped.  The method returns True if any of the
        remaining path components matches the ignore pattern.  In other words,
        if the path describes a file that is is in a directory that should be
        ignored it will also be ignored.

        The path can be relative or absolute.  Relative paths are *not* made
        absolute prior to comparison so this method is not affected by the
        current directory, even if the current diretory would itself be
        ignored."""
        fpath = fpath.normpath()
        while True:
            head, tail = fpath.split()
            if (tail and tail != fpath.pardir and
                    self.ignore_file(to_text(tail))):
                return True
            if not head or head == fpath:
                # No head left, or the path is unsplitable
                return False
            fpath = head

    @old_method('RebuildFileTable')
    def rebuild_file_table(self):
        """Rescans the file system and manifest and rebuilds the
        :py:attr:`fileTable`."""
        self.fileTable = {}
        been_there = {}
        for f in self.dPath.listdir():
            if self.ignore_file(to_text(f)):
                continue
            if f.normcase() == 'imsmanifest.xml':
                continue
            self.file_scanner(f, been_there)
        # Now scan the manifest and identify which file objects refer to which
        # files
        for r in self.manifest.root.Resources.Resource:
            for f in r.File:
                fpath = f.package_path(self)
                if fpath is None:
                    continue
                if fpath in self.fileTable:
                    self.fileTable[fpath].append(f)
                else:
                    self.fileTable[fpath] = [f]

    @old_method('FileScanner')
    def file_scanner(self, fpath, been_there):
        full_path = self.dPath.join(fpath)
        rfull_path = full_path.realpath()
        if rfull_path in been_there:
            raise CPPackageBeenThereError(rfull_path)
        been_there[rfull_path] = True
        if full_path.isdir():
            for f in full_path.listdir():
                if self.ignore_file(to_text(f)):
                    continue
                self.file_scanner(fpath.join(f), been_there)
        elif full_path.isfile():
            self.fileTable[fpath.normcase()] = []
        else:  # skip non-regular files.
            pass

    @old_method('PackagePath')
    def package_path(self, fpath):
        """Converts an absolute file path into a canonical package-relative path

        Returns None if fpath is not inside the package."""
        rel_path = []
        assert isinstance(fpath, vfs.VirtualFilePath)
        while fpath != self.dPath:
            fpath, tail = fpath.split()
            if not fpath or not tail:
                # We've gone as far as we can, fail!
                return None
            rel_path[0:0] = [tail]
        return self.dPath.__class__(*rel_path).normcase()

    @old_method('ExpandZip')
    def expand_zip(self, zf):
        self.dPath = vfs.defaultFS.mkdtemp('.d', 'imscpv1p2-')
        self.tempDir = True
        zf = zipfile.ZipFile(zf)
        try:
            for zfi in zf.infolist():
                path = self.dPath
                for path_seg in zfi.filename.split('/'):
                    # The current path will need to be a directory
                    if not path.isdir():
                        path.mkdir()
                    if isinstance(path_seg, bytes):
                        path_seg = path_seg.decode('utf-8')
                    elif not py2 and (zfi.flag_bits ^ 0x800):
                        # Python3 decodes as cp437 if the flag is clear
                        # we force utf-8 in this case
                        path_seg = path_seg.encode('cp437').decode('utf-8')
                    path = path.join(path_seg).normpath()
                    if self.package_path(path) is None:
                        raise CPZIPFilenameError(zfi.filename)
                if path.isdir():
                    if zfi.file_size > 0:
                        raise CPZIPDirectorySizeError(
                            "%s has size %i" % (zfi.filename, zfi.file_size))
                elif path.exists():
                    # Duplicate entries in the zip file
                    raise CPZIPDuplicateFileError(zfi.filename)
                else:
                    f = path.open('wb')
                    try:
                        f.write(zf.read(zfi.filename))
                    finally:
                        f.close()
        finally:
            zf.close()

    @old_method('ExportToPIF')
    def export_to_pif(self, zpath):
        """Exports the content package, saving the zipped package in *zpath*

        *zpath* is overwritten by this operation.

        In order to make content packages more interoperable this method goes
        beyond the basic zip specification and ensures that pathnames are
        always UTF-8 encoded when added to the archive.  When creating
        instances of :py:class:`ContentPackage` from an existing archive the
        reverse transformation is performed.  When exchanging PIF files between
        systems with different native file path encodings, encoding erros may
        occur."""
        zf = zipfile.ZipFile(zpath, 'w')
        base = b''
        been_there = {}
        try:
            for f in self.dPath.listdir():
                if self.ignore_file(to_text(f)):
                    continue
                self.add_to_zip(self.dPath.join(f), zf, base, been_there)
        finally:
            zf.close()

    @old_method('AddToZip')
    def add_to_zip(self, fpath, zf, zbase, been_there):
        rfname = fpath.realpath()
        if rfname in been_there:
            raise CPZIPBeenThereError(fpath)
        been_there[rfname] = True
        fname = to_text(fpath.split()[1])
        zfname = fname.replace('/', ':')
# 		if type(zfname) is StringType:
# 			zfname=zfname.decode(sys.getfilesystemencoding())
        zpath = zbase + zfname.encode('utf-8')
        if fpath.isdir():
            zpath += b'/'
            if py2:
                zf.writestr(zpath, b'')
            else:
                zf.writestr(zpath.decode('utf-8'), b'')
            for f in fpath.listdir():
                if self.ignore_file(to_text(f)):
                    continue
                self.add_to_zip(fpath.join(f), zf, zpath, been_there)
        elif fpath.isfile():
            with vfs.ZipHooks():
                # in Python3 zpath must be a string
                if not py2:
                    zpath = zpath.decode('utf-8')
                zf.write(fpath, zpath)
        else:  # skip non-regular files.
            pass

    @old_method('GetUniqueFile')
    def get_unique_path(self, suggested_path):
        """Returns a unique file path suitable for creating a new file in the package.

        suggested_path is used to provide a suggested path for the file.  This
        may be relative (to the root and manifest) or absolute but it must
        resolve to a file (potentially) in the package.  The suggested_path
        should either be a VirtualFilePath (of the same type as the content
        package's :py:attr:`dPath`) or a string suitable for conversion to a
        VirtualFilePath.

        When suggested_path is relative, it is forced to lower-case.  This is
        consistent with the behaviour of normcase on systems that are case
        insensitive.  The trouble with case insensitive file systems is that it
        may be impossible to unpack a content package created on a case
        sensitive system and store it on a case insenstive one.  By channelling
        all file storage through this method (and constructing any URIs *after*
        the file has been stored) the resulting packages will be more portable.

        If suggested_path already corresponds to a file already in the package,
        or to a file already referred to in the manifest, then a random string
        is added to it while preserving the suggested extension in order to
        make it unique.

        The return result is always normalized and returned relative to the
        package root."""
        if not isinstance(suggested_path, vfs.VirtualFilePath):
            suggested_path = self.dPath.__class__(suggested_path)
        if suggested_path.isabs():
            fpath = suggested_path
        else:
            fpath = self.dPath.join(to_text(suggested_path).lower())
        fpath = path_in_path(fpath, self.dPath)
        if fpath is None:
            raise CPFilePathError(suggested_path)
        fpath = fpath.normcase()
        # Now we can try and make it unique
        path_str = fpath
        path_extra = 0
        while path_str in self.fileTable:
            if not path_extra:
                path_extra = random.randint(0, 0xFFFF)
            fname, fExt = fpath.splitext()
            path_str = '%s_%X%s' % (to_text(fname), path_extra, fExt)
            path_extra = path_extra + 1
        # we have the path string
        return self.dPath.__class__(path_str)

    @old_method('File')
    def new_file(self, resource, href):
        """Returns a new :py:class:`File` object attached to *resource*

        *href* is the URI of the file expressed relative to the resource
        element in the manifest.  Although this is normally the same as the URI
        expressed relative to the package, a resource may have an xml:base
        attribute that alters the base for resolving relative URIs.

        *href* may of course be an absolute URI to an external resource.  If an
        absolute URI is given to a local file it must be located inside the
        package.

        Attempting to add a :py:class:`File` object representing the manifest
        file iteself will raise :py:class:`CPFilePathError`.

        The :py:attr:`fileTable` is updated automatically by this method."""
        furl = resource.resolve_uri(href)
        if not isinstance(furl, uri.FileURL):
            # Not a local file
            f = resource.add_child(resource.FileClass)
            f.href = href
        else:
            if href.is_absolute():
                href = uri.URIFactory.relative(href, resource.resolve_base())
            full_path = furl.get_virtual_file_path()
            head, tail = full_path.split()
            if self.ignore_file(to_text(tail)):
                raise CPFilePathError(full_path)
            rel_path = path_in_path(full_path, self.dPath)
            if (rel_path is None or
                    to_text(rel_path).lower() == 'imsmanifest.xml'):
                raise CPFilePathError(furl.path)
            # normalise the case ready to put in the file table
            rel_path = rel_path.normcase()
            f = resource.add_child(resource.FileClass)
            f.href = href
            if rel_path not in self.fileTable:
                self.fileTable[rel_path] = [f]
            else:
                self.fileTable[rel_path].append(f)
        return f

    @old_method('FileCopy')
    def file_copy(self, resource, src_url):
        """Returns a new :py:class:`File` object copied into the package from
        *src_url*, attached to *resource*.

        The file is copied to the same directory as the resource's entry point
        or to the main package directory if the resource has no entry point.

        The :py:class:`File` object is actually created with the
        :py:meth:`File` method.

        Note that if src_url points to a missing file then no file is copied to
        the package but the associated :py:class:`File` is still created.
        It will point to a missing file."""
        src_path = src_url.get_virtual_file_path()
        # We need to create a new file object
        fstart = resource.get_entry_point()
        if fstart is None:
            base_path = self.dPath
        else:
            url = fstart.resolve_uri(fstart.href)
            if not isinstance(url, uri.FileURL):
                base_path = self.dPath
            else:
                base_path, tail = url.get_virtual_file_path().split()
        # now pick up the last component of src
        head, tail = src_path.split()
        new_src_path = self.get_unique_path(base_path.join(tail))
        new_src_path = self.dPath.join(new_src_path)
        new_src = uri.URI.from_virtual_path(new_src_path)
        # Turn this file path into a relative URL in the context of the new
        # resource
        href = resource.relative_uri(new_src)
        f = self.File(resource, href)
        dname, fname = new_src_path.split()
        if not dname.isdir():
            dname.makedirs()
        if src_path.isfile():
            src_path.copy(new_src_path)
        return f

    @old_method('DeleteFile')
    def delete_file(self, href):
        """Removes the file at *href* from the file system

        This method also removes any file references to it from resources in
        the manifest. href may be given relative to the package root directory.
        The entry in :py:attr:`fileTable` is also removed.

        :py:class:`CPFileTypeError` is raised if the file is not a regular file

        :py:class:`CPFilePathError` is raised if the file is an
        :py:meth:`ignore_file`, the manifest itself or outside of the content
        package.

        :py:class:`CPProtocolError` is raised if the indicated file is not in
        the local file system."""
        base_uri = self.manifest.get_base()
        base = uri.URI.from_octets(base_uri)
        furl = uri.URI.from_octets(href).resolve(base)
        if not isinstance(furl, uri.FileURL):
            # We cannot delete non-file objects (though in future
            # we should support HTTP DELETE)
            return CPProtocolError(str(furl))
        full_path = furl.get_virtual_file_path()
        if not full_path.isfile():
            raise CPFileTypeError(full_path)
        head, tail = full_path.split()
        if self.ignore_file(to_text(tail)):
            raise CPFilePathError(full_path)
        rel_path = path_in_path(full_path, self.dPath)
        if rel_path is None or to_text(rel_path).lower() == 'imsmanifest.xml':
            raise CPFilePathError(full_path)
        # normalise the case ready for comparisons
        rel_path = rel_path.normcase()
        for r in self.manifest.root.Resources.Resource:
            del_list = []
            for f in r.File:
                # Does f point to the same file?
                if f.package_path(self) == rel_path:
                    del_list.append(f)
            for f in del_list:
                r.delete_file(f)
        # Now there are no more references, safe to remove the file itself
        full_path.remove()
        if rel_path in self.fileTable:
            del self.fileTable[rel_path]

    @old_method('GetPackageName')
    def get_package_name(self):
        """Returns a human readable name for the package

        The name is determined by the method used to create the object. The
        purpose is to return a name that would be intuitive to the user if it
        were to be used as the name of the package directory or the stem of a
        file name when exporting to a PIF file.

        Note that the name is returned as a unicode string suitable for showing
        to the user and may need to be encoded before being used in file path
        operations."""
        if isinstance(self.packageName, bytes):
            return self.packageName.decode('utf-8')
        else:
            return to_text(self.packageName)

    @old_method('Close')
    def close(self):
        """Closes the content package, removing any temporary files.

        This method must be called to clean up any temporary files created when
        processing the content package.  Temporary files are created inside a
        special temporary directory created using the builtin python
        tempdir.mkdtemp function.  They are not automatically cleaned up when
        the process exits or when the garbage collector disposes of the object.
        Use of try:... finally: to clean up the package is recommended.  For
        example::

                pkg=ContentPackage("MyPackage.zip")
                try:
                        # do stuff with the content package here
                finally:
                        pkg.Close()

        """
        self.manifest = None
        self.fileTable = {}
        if self.tempDir and self.dPath:
            self.dPath.rmtree(True)
            self.dPath = None



#! /usr/bin/env python

from . import common
from ..xml import structures as xml


class ObjectBank(common.MetadataContainerMixin, common.QTICommentContainer):

    """This is the container for the Section(s) and/or Item(s) that are to be
    grouped as an object-bank. The object-bank is assigned its own unique
    identifier and can have the full set of QTI-specific meta-data::

    <!ELEMENT objectbank (qticomment? , qtimetadata* , (section | item)+)>
    <!ATTLIST objectbank  ident CDATA  #REQUIRED >"""
    XMLNAME = "objectbank"
    XMLATTR_ident = 'ident'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        common.QTICommentContainer.__init__(self, parent)
        self.ident = None
        self.QTIMetadata = []
        self.ObjectMixin = []

    def migrate_to_v2(self, output):
        for sectionOrItem in self.ObjectMixin:
            sectionOrItem.migrate_to_v2(output)



#! /usr/bin/env python

import errno
import io
import logging
import os
import select
import socket
import ssl
import threading
import time

from .. import rfc2396 as uri
from ..py2 import py2, dict_items
from ..py26 import *       # noqa
from ..streams import Pipe

from . import grammar, messages, params

if py2:
    import Queue as queue
    import SocketServer as socketserver
    from threading import _Event as EventClass
else:
    import queue
    import socketserver
    from threading import Event as EventClass


#: For useful information on HTTP header size limits see:
# : http://stackoverflow.com/questions/1097651/is-there-a-practical-http-header-length-limit     # noqa
#: we match Apache's value by default
MAX_HEADER_LINE = 8190

#: we match Apache's value by default
MAX_HEADER_SIZE = 2 * MAX_HEADER_LINE


class UnresponsiveScript(messages.HTTPException):
    pass


def cascading_wait(timeout, *events):
    wait_event = threading.Event()
    for e in events:
        e.cascade_to(wait_event)
    wait_event.wait(timeout)


class CascadingEvent(EventClass):

    def __init__(self):
        super(CascadingEvent, self).__init__()
        self.cascade = None

    def cascade_to(self, e):
        """Instructs this event to cascade to e"""
        self.cascade = e
        # at this point, set will cascade, but if we
        # were set before we need to cascade now
        if self.is_set():
            e.set()

    def set(self):
        super(CascadingEvent, self).set()
        e = self.cascade
        if e is not None:
            e.set()


def split_socket(s, buffstr=b'', timeout=None,
                 maxline=MAX_HEADER_LINE,
                 maxlines=MAX_HEADER_SIZE):
    """Function to help reading from non-blocking sockets.

    s
        A socket object

    buffstr
        An optional string object containing buffered data returned
        from a previous call.

    timeout
        The length of time in seconds to wait for data on the
        socket. None indicates wait forever.

    maxline
        The maximum number of bytes to allow in a single header
        line, defaults to :py:data:`MAX_HEADER_LINE`.  Lines that
        exceed this value raise an :py:class:`HTTPException`.

    maxlines
        The maximum number of bytes to allow for all lines combined,
        defaults to :py:data:`MAX_HEADER_SIZE`.  Lines that exceed
        this total raise an :py:class:`HTTPException`.

    Returns a tuple of two items.  The first item is a list of
    line-strings read from the socket.  Each line string is
    terminated by CRLF, the last string in the list is a blank line
    consisting of CRLF only.  Therefore, the list is always
    non-empty on a successful return.

    The second item is any extra data read from the socket after the
    blank line.  It is always a string, but may be empty."""
    buffer = bytearray(buffstr)
    rpos = 0
    lines = []
    while True:
        # scan up to the next CRLF in buffer
        try:
            end = buffer.index(grammar.CRLF, rpos)
            line = bytes(buffer[rpos:end + 2])
            if len(line) > maxline:
                raise messages.HTTPException(
                    "max line length exceeded: %s..." + line[0:64])
            lines.append(line)
            rpos = end + 2
            if line == grammar.CRLF:
                # that was the last line
                return lines, bytes(buffer[rpos:])
            else:
                # read another line
                continue
        except ValueError:
            # not found, fill the buffer with data
            pass
        if len(buffer) >= maxlines:
            raise messages.HTTPException("max header length exceeded")
        data = read_socket(s, maxlines - len(buffer), timeout=timeout)
        if not data:
            raise messages.ProtocolError("Unexpected end of message")
        buffer = buffer + data


def split_socket1(s, buffstr=b'', timeout=None, maxline=MAX_HEADER_LINE):
    """Function to help reading from non-blocking sockets.

    s
        A socket object

    buffstr
        An optional string object containing buffered data returned
        from a previous call.

    timeout
        The length of time in seconds to wait for data on the
        socket. None indicates wait forever.

    maxline
        The maximum number of bytes to allow in a single header
        line, defaults to :py:data:`MAX_HEADER_LINE`.  Lines that
        exceed this value raise an :py:class:`HTTPException`.

    Returns a tuple of two items.  The first item is a string
    terminating in CRLF, the second item is any extra data read from the
    socket after the blank line.  It is always a string, but may be
    empty.

    If the other end of the socket has closed before a complete line was
    sent then the first item is None and the second is any data read
    from the socket before it was closed."""
    buffer = bytearray(buffstr)
    rpos = 0
    while True:
        # scan up to the next CRLF in buffer
        try:
            end = buffer.index(grammar.CRLF, rpos)
            line = bytes(buffer[rpos:end + 2])
            rpos = end + 2
            return line, bytes(buffer[rpos:])
        except ValueError:
            # not found, fill the buffer with data
            pass
        if len(buffer) >= maxline:
            raise HTTPException("max line length exceeded: %s..." +
                                buffer[0:64])
        data = read_socket(s, maxline - len(buffer), timeout=timeout)
        if not data:
            return None, bytes(buffer)
        buffer = buffer + data


def read_socket(s, nbytes, timeout=None):
    """Function to help reading from non-blocking sockets.

    s
        A socket object

    nbytes
        The maximum number of bytes to read.  The function will block
        attempting to read from the socket unless at least 1 byte is
        returned.

    timeout
        The length of time in seconds to wait for data on the
        socket. None indicates wait forever.

    This method has three outcomes, a data string is returned, None is
    returned (client hung up) or an IOError is raised indicating a
    timeout."""
    try:
        r, w, e = select.select([s], [], [], timeout)
        if not r:
            logging.info("socket timeout on recv")
            raise IOError(errno.ETIMEDOUT, os.strerror(errno.ETIMEDOUT),
                          "select in pyslet.http.messages.read_socket")
    except select.error as err:
        logging.error("Socket error from select: %s", str(err))
        raise IOError(err.args[0], err.args[1])
    try:
        data = s.recv(nbytes)
        # if data is empty the other end has hung up
        if data == b'' and nbytes > 0:
            return None
        return data
    except socket.error:
        # error raised on recv ready socket indicates a hang-up
        # typically: i.e., no more data to come
        logging.info("Error raised on recv ready socket")
        return None


def write_socket(s, data, timeout=None):
    """Function to help writing to a non-blocking socket

    s
        A socket object

    data
        The data to write.  The function will block attempting to write
        to the socket until all the data has been written or an error is
        raised.

    timeout
        The length of time in seconds to wait for the socket to be ready
        to write.  None indicates wait forever."""
    while data:
        try:
            r, w, e = select.select([], [s], [], timeout)
            if not w:
                logging.info("socket timeout on send")
                raise IOError(errno.ETIMEDOUT, os.strerror(errno.ETIMEDOUT),
                              "select in pyslet.http.messages.write_socket")
        except select.error as err:
            logging.error("Socket error from select: %s", str(err))
            raise IOError(err.args[0], err.args[1])
        try:
            nbytes = s.send(data)
            if nbytes:
                if nbytes < len(data):
                    data = data[nbytes:]
                else:
                    data = None
            else:
                # socket was ready to write but didn't accept any data
                # more typically represented by socket.error caught below
                logging.info("Send failed on send-ready socket")
                raise IOError(errno.EPIPE, os.strerror(errno.EPIPE))
        except socket.error:
            # error raised on recv ready socket indicates a hang-up
            # typically: i.e., no more data to come
            logging.info("Error raised on send-ready socket")
            return None


class Connection(socketserver.BaseRequestHandler):

    def handle(self):
        self.server.handle_connection(self)

    def begin(self):
        # a FIFO queue of responses
        self.request.setblocking(False)
        self.responseq = queue.Queue()
        self.finished = threading.Event()

    def handle_requests(self):
        # these pipes are for the entity body, not the message body! in
        # otherwords, they deal with data after transfer encodings have
        # been removed
        try:
            while not self.finished.is_set():
                logging.debug("handle_request: reading request...")
                request = ServerRequest(self)
                rflag = CascadingEvent()
                request.recv_pipe.set_rflag(rflag)
                request.start_receiving()
                buffstr = b''
                timeout = self.server.idle_timeout
                send_continue = False
                send_continue_start = 0
                try:
                    while True:
                        mode = request.recv_mode()
                        if mode is None or request.aborted.is_set():
                            # the request is complete, simulate EOF on
                            # the recv_pipe to ensure that any read()
                            # call in the application terminates and
                            # doesn't just hang
                            logging.debug("%s : request complete",
                                          request.get_start())
                            request.recv_pipe.write_eof()
                            break
                        if send_continue:
                            if not rflag.is_set():
                                twait = (self.server.app_timeout -
                                         (time.time() - send_continue_start))
                                if twait < 0:
                                    # application timeout
                                    raise UnresponsiveScript(
                                        "timed out waiting to consume "
                                        "input data")
                                else:
                                    # wait for a reader before continuing
                                    cascading_wait(
                                        twait, rflag, request.aborted)
                                    continue
                            else:
                                # script is read blocked, send 100-Continue
                                request.response.send_continue.set()
                            send_continue = False
                        if mode == request.RECV_HEADERS:
                            headers, buffstr = split_socket(
                                self.request, buffstr, timeout=timeout)
                            request.recv(headers)
                            # at this point we should have a response
                            if request.get_expect_continue():
                                send_continue = True
                                send_continue_start = time.time()
                        elif mode == request.RECV_LINE:
                            # we need to read a line from the input stream
                            line, buffstr = split_socket1(
                                self.request, buffstr, timeout=timeout)
                            if line is None:
                                # hang up
                                if request.method:
                                    raise ProtocolError("Unexpected EOM")
                                else:
                                    # client hung up
                                    logging.debug("client hang up detected")
                                    self.finished.set()
                                    break
                            else:
                                request.recv(line)
                        elif mode > 0:
                            # we need to read some data from the socket
                            if len(buffstr) > mode:
                                data = buffstr[:mode]
                                buffstr = buffstr[mode:]
                                request.recv(data)
                            elif buffstr:
                                request.recv(buffstr)
                                buffstr = ''
                            else:
                                # read some data from the socket
                                data = read_socket(self.request, mode,
                                                   timeout=timeout)
                                if data is None:
                                    # client hang up when we expected
                                    # more data
                                    raise ProtocolError(
                                        "Unexpected EOM in data")
                                request.recv(data)
                        elif mode == request.RECV_ALL:
                            # unlimited read from the socket
                            if buffstr:
                                request.recv(buffstr)
                                buffstr = ''
                            else:
                                data = read_socket(self.request,
                                                   io.DEFAULT_BUFFER_SIZE,
                                                   timeout=timeout)
                                request.recv(data)
                                if not data:
                                    self.finished.set()
                        elif mode == 0:
                            # the request is write blocked, it doesn't
                            # want any data, it just wants time to
                            # digest what it has already.  Typically we
                            # get here when a POST or similar method is
                            # waiting for the script to read the last
                            # bit of the data.  It makes sense to wait
                            # for the script to finish processing which
                            # means waiting for a read on the recv_pipe
                            if rflag.is_set():
                                rflag.clear()
                                # next time around we'll wait for it
                            else:
                                logging.debug(
                                    "handle_request: recv_pipe blocked, "
                                    "waiting for app")
                                cascading_wait(self.server.app_timeout,
                                               rflag, request.aborted)
                                if not rflag.is_set():
                                    raise UnresponsiveScript(
                                        "timed out waiting for app to consume "
                                        "input data")
                            # give time to the message
                            request.recv(None)
                            pass
                        # once we're reading a request, switch to a longer
                        # timeout to prevent unexpected hang-ups
                        timeout = self.server.connection_timeout
                except messages.ProtocolError as e:
                    # log this error as a warning
                    # push a suitable response and then
                    # hang up, no good will come of this connection
                    response = ServerResponse(request=request,
                                              protocol=self.server.protocol)
                    self.responseq.put(response)
                    response.set_status(400)
                    response.clear_keep_alive()
                    txt = str(e).encode('iso8859-1')
                    response.set_content_type(params.PLAIN_TEXT)
                    response.set_content_length(len(txt))
                    response.write_response(txt)
                    self.finished.set()
                except NotImplementedError as e:
                    response = ServerResponse(request=request,
                                              protocol=self.server.protocol)
                    self.responseq.put(response)
                    response.set_status(501)
                    response.clear_keep_alive()
                    txt = str(e).encode('iso8859-1')
                    response.set_content_type(params.PLAIN_TEXT)
                    response.set_content_length(len(txt))
                    response.write_response(txt)
                    self.finished.set()
                except IOError as e:
                    if request.method:
                        logging.warning("Unexpected EOM: %s; %s",
                                        request.get_start(), str(e))
                    else:
                        # client may have hung up
                        logging.debug("timed out waiting for request")
                    request.clear_keep_alive()
                    self.finished.set()
                except Exception as e:
                    # log this error as an internal server error push a
                    # suitable 500 response and then hang up because
                    # we've messed up
                    response = ServerResponse(request=request,
                                              protocol=self.server.protocol)
                    self.responseq.put(response)
                    response.set_status(500)
                    response.clear_keep_alive()
                    txt = str(e).encode('iso8859-1')
                    # import traceback
                    # txt = traceback.format_exc().encode('iso-8859-1')
                    logging.error("%s: %s", response.request.get_start(), txt)
                    response.set_content_type(params.PLAIN_TEXT)
                    response.set_content_length(len(txt))
                    response.write_response(txt)
                    self.finished.set()
                # we've finished reading the message, close the recv_pipe
                # to help clean things up quickly
                request.recv_pipe.close()
                if not request.keep_alive or request.response is None:
                    # the request wants us to close the connection or
                    # the client has hung up (or something has gone
                    # badly wrong)
                    break
                # wait for the response to be ready to send before
                # pipe-lining the next request in case it wants to
                # hang up the connection.
                logging.debug("handle_request: waiting for response...")
                request.response.ready_to_send.wait(self.server.app_timeout)
                if not request.response.keep_alive:
                    # the response is terminating the connection
                    break
                # request and response happy to keep the connection open
                continue
        except IOError:
            # socket may have timed-out while reading or waiting for a
            # request
            pass
        # add a dummy response to kill the handle_responses thread
        logging.debug("handle_request: terminating connection...")
        self.responseq.put(None)

    def handle_response(self, response):
        self.responseq.put(response)

    def handle_responses(self):
        # pull items from the responseq until we get None
        timeout = (self.server.connection_timeout +
                   self.server.app_timeout)
        try:
            keep_alive = True
            while keep_alive:
                logging.debug("handle_responses: waiting for responseq...")
                response = self.responseq.get(True, timeout)
                if response is None:
                    break
                # handle the response
                logging.debug("Waiting for response...")
                # triggered by a call to response.start_sending()
                # or the need for a 100-Continue
                cascading_wait(self.server.app_timeout,
                               response.ready_to_send,
                               response.send_continue)
                if response.request.get_expect_continue():
                    # ignore the semaphore, its purpose was purely
                    # to wake us up early
                    if (not response.ready_to_send.is_set() or
                            response.status is None or
                            (response.status >= 200 and
                             response.status < 400)):
                        logging.info("Sending response: %s %s",
                                     str(response.protocol), "100 Continue")
                        # Fake this response, as the app may already be
                        # setting the status, headers etc...
                        data = b"%s %s\r\n\r\n" % (
                            response.protocol.to_bytes(), b"100 Continue")
                        write_socket(self.request, data,
                                     self.server.connection_timeout)
                        # now wait again until we're actually ready to send
                        response.ready_to_send.wait(self.server.app_timeout)
                    else:
                        # ready to send and status does not indicate
                        # success send a final code, abort reading the
                        # data and hang up the connection
                        if response.keep_alive:
                            # we missed this opportunity before
                            response.clear_keep_alive()
                            response.set_connection(["close"])
                        response.request.aborted.set()
                if not response.ready_to_send.is_set():
                    # script timeout: generate a 500 error then kill
                    # this connection
                    response.set_status(500)
                    response.clear_keep_alive()
                    # we can't risk writing data as a race condition
                    # could leave us dead-locked waiting for the loop
                    # below (that clears the response Pipe)
                    logging.error("%s: Application timeout",
                                  response.request.get_start())
                    response.set_content_length(0)
                    response.write_response('')
                logging.debug("...sending response")
                # send the response
                keep_alive = response.keep_alive
                data = response.send_start()
                write_socket(self.request, data,
                             self.server.connection_timeout)
                # now read the headers and the rest of the message
                data = response.send_header()
                write_socket(self.request, data,
                             self.server.connection_timeout)
                # now loop round reading the data
                while True:
                    data = response.send_body()
                    if data is None:
                        # we're read blocked, wait for the response's
                        # pipe to be readable
                        response.send_pipe.read_wait(
                            timeout=self.server.app_timeout)
                    elif data:
                        write_socket(self.request, data,
                                     self.server.connection_timeout)
                    else:
                        # end of body, close the response pipe
                        response.close()
                        break
        except queue.Empty:
            # timeout waiting for a response object
            logging.error("HTTPServer: responseq timeout "
                          "(may indicate a stuck connection)")
        self.finished.set()
        logging.debug("handle_responses: done")

    def end(self):
        pass

    def service_unavailable(self):
        # return a 503 response
        write_socket(self.request,
                     "HTTP/1.1 503 Service Unavailable\r\n"
                     "Connection: close\r\n"
                     "\r\n",
                     self.server.connection_timeout)


class Server(socketserver.ThreadingMixIn, socketserver.TCPServer):

    """HTTP Server

    The purpose of this class is to provide a simple environment for
    testing and customising specialist aspects of HTTP handling.

    port
        The port on which to listen for connections

    max_connections
        The maximum number of connections to process at any one time.
        Connections are placed in a queue and processed using a round
        robin algorithm by a separate thread.  None, the default, means
        unlimited.

    app
        A wsgi callable that will handle requests

    protocol
        A :py:class:`params.HTTPVersion` instance representing the
        protocol implemented by the server.  The default is HTTP/1.1, if
        you specify :py:data:`params.HTTP_1p0` the server will behave
        more like an HTTP/1.0 server, reporting that version in
        responses.  It also suppresses the use of chunked encoding and
        persistent connections.

    authorities
        A list of strings representing the authorities for which we are
        serving.  The first authority in the list is considered the
        default authority and it will be used in cases where the request
        does not provide a host header (or absolute URI).  If the list
        is empty or missing (the default) then the default authority is
        set by adding the port number to the string "localhost:".  Use
        multiple names to enable the server to recognise other
        authorities, any unrecognized authority received in a request
        will return a 400 bad request message.

        The default authority is also used to determine the host string
        passed to the underlying system for binding the server.  The
        port given in the authority string is ignored when binding but
        in most cases it should correspond to the port passed as the
        first parmeter!

    keyfile, certfile
        Optional paths to an SSL key and certificate file.  If given the
        server is an https server.  For an explanation of these
        arguments see the builtin Python ssl.wrap_socket function to
        which they are passed."""
    #: overridden to allow our server to restart even if there are
    #: existing connections from a previous invocation.
    allow_reuse_address = True

    def __init__(self, port, max_connections=None, app=None,
                 protocol=params.HTTP_1p1, authorities=None,
                 keyfile=None, certfile=None):
        #: a dictionary mapping authority onto the WSGI callable that
        #: handles it
        self.authorities = {}
        if not authorities:
            #: the HOST we are bound to
            authorities = ["localhost" if port == params.HTTPURL.DEFAULT_PORT
                           else "localhost:%i" % port]
        #: the default authority
        self.default_authority = authorities[0].lower()
        self.host = uri.split_server(self.default_authority)[1]
        for a in authorities:
            self.authorities[a.lower()] = app
        #: the port we are bound to
        self.port = port
        #: whether or not we are serving using https
        self.https = False
        socketserver.TCPServer.__init__(
            self, (self.host, self.port), Connection)
        if keyfile is not None:
            # This must be an HTTPS server
            self.socket = ssl.wrap_socket(
                self.socket, keyfile=keyfile, certfile=certfile,
                server_side=True, do_handshake_on_connect=True)
            self.https = True
        #: the protocol semantics we use for handling requests
        self.protocol = protocol
        self.lock = threading.RLock()
        self.con_count = 0
        self.con_max = max_connections
        self.nconnections = 0
        self.connections = {}
        #: seconds before a network recv/send times out
        self.connection_timeout = 10
        #: seconds before a working request handler is abandoned
        self.app_timeout = 10
        #: seconds before an idle connection is closed
        self.idle_timeout = 5
        #: a pipe for writing error strings
        self.error_pipe = Pipe(rblocking=False, timeout=self.app_timeout,
                               name="http.Server.error_pipe")

    def handle_connection(self, connection):
        # add this connection to the queue
        with self.lock:
            if (self.con_max is not None and
                    len(self.connections) > self.con_max):
                # refuse this connection, we're too busy
                connection.id = 0
            else:
                self.con_count += 1
                connection.id = self.con_count
                self.connections[connection.id] = connection
        if connection.id:
            connection.begin()
            t = threading.Thread(target=connection.handle_responses)
            t.start()
            connection.handle_requests()
            # now when we are done handling requests we just wait
            # for the responses thread to finish
            t.join()
            # end of this connection
            connection.end()
            self.stop_connection(connection)
        else:
            connection.service_unavailable()

    def stop_connection(self, connection):
        with self.lock:
            if connection.id and connection.id in self.connections:
                del self.connections[connection.id]

    def launch_app(self, environ, start_response):
        """Launches a WSGI application

        The default implementation returns a 404 error"""
        with self.lock:
            authority = environ.get('HTTP_HOST', self.default_authority)
            try:
                app = self.authorities[authority]
                if app is None:
                    # this implementation mainly for testing, we want to read
                    # all the data from the input string
                    response_headers = []
                    start_response("404 Page Not Found", response_headers)
                    return []
                else:
                    return app(environ, start_response)
            except KeyError:
                # bad host in request
                start_response("400 Bad Request", [])
                return []

    def get_request(self):
        """Adds logging of socket errors on connect

        The default implementations simply swallows socket errors making
        it hard to see why calls like handle_request have simply
        returned without actually doing anything."""
        try:
            return socketserver.TCPServer.get_request(self)
        except socket.error as e:
            logging.errror("Server socket error %s", repr(e))
            raise

    def handle_timeout(self):
        """Adds logging for timeouts during handle_request"""
        logging.error("HTTP Server timeout: %s" % repr(self.timeout))


class ServerRequest(messages.Request):

    def __init__(self, connection, **kwargs):
        self.connection = connection
        # the body won't block on write, but will block on read for up
        # to connection_timeout (as the data is coming from outside).
        self.recv_pipe = Pipe(wblocking=False,
                              timeout=connection.server.connection_timeout,
                              name="ServerRequest[%i].recv_pipe" %
                              self.connection.id)
        self.aborted = CascadingEvent()
        super(ServerRequest, self).__init__(entity_body=self.recv_pipe,
                                            **kwargs)

    def handle_headers(self):
        """Create a wsgi environment and start a processing thread"""
        # normalise the request URI to remove scheme and authority
        # ensures Host: is set correctly
        self.extract_authority()
        with self.lock:
            url = uri.URI.from_octets(self.request_uri)
            try:
                path_info = uri.unescape_data(url.abs_path).decode('utf-8')
            except UnicodeDecodeError:
                # ok, not UTF-8 then, try iso-8859-1
                path_info = uri.unescape_data(
                    url.abs_path).decode('iso-8859-1')
            environ = {
                'REQUEST_METHOD': self.method,
                'SCRIPT_NAME': '',
                'PATH_INFO': path_info,
                'QUERY_STRING': url.query,
                'SERVER_NAME': self.connection.server.host,
                'SERVER_PORT': str(self.connection.server.port),
                'SERVER_PROTCOL': str(self.protocol),
                'REMOTE_ADDR': str(self.connection.client_address),
                'wsgi.version': (1, 0),
                'wsgi.url_scheme':
                'https' if self.connection.server.https else 'http',
                'wsgi.input': io.BufferedReader(self.recv_pipe),
                'wsgi.errors': self.connection.server.error_pipe,
                'wsgi.multithread': True,
                'wsgi.multiprocess': False,
                'wsgi.run_once': False}
            if url.query is not None:
                environ['QUERY_STRING'] = url.query
            content_type = self.get_header('Content-Type')
            if content_type is not None:
                environ['CONTENT_TYPE'] = content_type
            content_length = self.get_header('Content-Length')
            if content_length is not None:
                environ['CONTENT_LENGTH'] = content_length
            for hname, hvalue in dict_items(self.headers):
                hname = hname.replace(b"-", b"_").decode('iso-8859-1')
                environ['HTTP_' + hname.upper()] = hvalue[1].decode(
                    'iso-8859-1')
        # now launch the application in a separate thread
        response = ServerResponse(
            request=self, protocol=self.connection.server.protocol)
        self.connection.handle_response(response)
        t = threading.Thread(target=response.launch_app, args=(environ, ))
        t.start()


class ServerResponse(messages.Response):

    def __init__(self, request, **kwargs):
        self.connection = request.connection
        # the body won't block on read, but will block on write for up
        # to connection_timeout (as the data is being sent outside).
        self.send_pipe = Pipe(
            rblocking=False,
            timeout=request.connection.server.connection_timeout,
            name="ServerResponse[%i].send_pipe" % self.connection.id)
        #: a CascadingEvent set when the response headers are sent
        self.ready_to_send = CascadingEvent()
        self.send_continue = CascadingEvent()
        super(ServerResponse, self).__init__(
            request=request, entity_body=self.send_pipe, **kwargs)

    def start_sending(self):
        super(ServerResponse, self).start_sending()
        self.ready_to_send.set()

    def launch_app(self, environ):
        logging.debug("Calling wsgi application...")
        data = self.connection.server.launch_app(environ, self.start_response)
        # does data support len?
        try:
            datalen = len(data)
        except TypeError:
            # length is unknown, probably a generator function
            datalen = None
        for item in data:
            if not isinstance(item, bytes):
                logging.error("WSGI application yielded non bytes: %s",
                              repr(item))
            if datalen == 1:
                # this is all the data, set the body_len
                self.body_len = len(item)
            if not self.ready_to_send.is_set():
                # don't send the headers until the first bit of data
                # has been yielded
                self.start_sending()
            wbytes = item
            while wbytes:
                nbytes = self.send_pipe.write(wbytes)
                if nbytes < len(wbytes):
                    wbytes = wbytes[nbytes:]
                else:
                    wbytes = None
            # self.send_pipe.flush()
            # I considered using flush here, given the following
            # requirement: "WSGI servers, gateways, and middleware must
            # not delay the transmission of any block". But we have a
            # separate thread for Connection.handle_responses and this
            # is an acceptable solution: "Use a different thread to
            # ensure that the block continues to be transmitted while
            # the application produces the next block"
            #
            # There's one caveat here though, if a transfer encoding is
            # in effect it is possible that a small number of the bytes
            # written to the send_pipe will be stuck in the codec
            # waiting for the next block (or a final flush).  Rather
            # than conclude that WSGI is incompatible with transparent
            # transport encodings we're going to ignore this issue.
        if not self.ready_to_send.is_set():
            # empty response, send it now
            self.body_len = 0
            self.start_sending()
        # when we are done writing data to the send_pipe we tell the
        # response that there is no more in case the body length was
        # indeterminate and the reader is reading forever
        self.send_pipe.write_eof()
        # HTTP requires us to read the entire input pipe even if
        # we didn't use all the data, so do that here.
        input = environ['wsgi.input']
        # we are limited in the methods we are allowed to use here
        # except that we know it is a BufferedReader over a Pipe object
        # set for read blocking up to the connection timeout
        spool_count = 0
        while not self.request.aborted.is_set():
            # this loop monitors the aborted flag for safety. In
            # practice the recv thread polices any bad clients (such as
            # those engaged in DoS activity) and if it aborts the
            # request it will also generate an EOF on the input stream.
            # We'll stop at whichever is detected first!
            logging.debug("reading trailing data...")
            data = input.read1(io.DEFAULT_BUFFER_SIZE)
            if data:
                if spool_count < io.DEFAULT_BUFFER_SIZE:
                    logging.warning(
                        "wsgi application discarded %i bytes of data",
                        len(data))
                    logging.debug("discarding data: \n%s", data)
                spool_count += len(data)
            else:
                break

    def start_response(self, status, response_headers, exc_info=None):
        with self.lock:
            pstatus = params.ParameterParser(status, ignore_sp=False)
            if pstatus.is_integer():
                self.status = pstatus.parse_integer()
            else:
                self.status = 0
            pstatus.parse_sp()
            self.reason = pstatus.parse_remainder().decode('iso-8859-1')
            for name, value in response_headers:
                self.set_header(name, value)
        return self.write_response

    def write_response(self, body_data):
        if not self.ready_to_send.is_set():
            # start sending the response, we've got legacy data
            self.start_sending()
        while body_data:
            nbytes = self.send_pipe.write(body_data)
            if nbytes < len(body_data):
                body_data = body_data[nbytes:]
            else:
                body_data = None
            # see comment above, we don't need to flush
            # self.send_pipe.flush()

    def close(self):
        self.send_pipe.close()



#! /usr/bin/env python

import warnings

from pyslet.xml import structures, parser   # noqa

warnings.warn(
    "pyslet.xml20081126 is deprecated, use pyslet.xml.structures "
    "or pyslet.xml.parser instead", DeprecationWarning, stacklevel=3)



#! /usr/bin/env python
"""This module implements the EDMX specification defined by Microsoft.

http://msdn.microsoft.com/en-us/library/dd541284(v=prot.10)"""

import itertools

from . import csdl as edm
from .. import rfc2396 as uri
from ..xml import structures as xml
from ..xml import namespace as xmlns


#: Namespace to use for EDMX elements
EDMX_NAMESPACE = "http://schemas.microsoft.com/ado/2007/06/edmx"


class EDMXElement(xmlns.XMLNSElement):
    XMLCONTENT = xml.ElementType.ElementContent


class DataServices(edm.NameTableMixin, EDMXElement):
    XMLNAME = (EDMX_NAMESPACE, 'DataServices')

    def __init__(self, parent):
        EDMXElement.__init__(self, parent)
        edm.NameTableMixin.__init__(self)
        self.Schema = []

    def get_children(self):
        for s in self.Schema:
            yield s
        for child in super(DataServices, self).get_children():
            yield child

    def content_changed(self):
        for s in self.Schema:
            self.declare(s)
        for s in self.Schema:
            s.update_type_refs(self)
        for s in self.Schema:
            s.update_set_refs(self)

    def validate(self):
        for s in self.Schema:
            s.validate()


class Reference(EDMXElement):
    XMLNAME = (EDMX_NAMESPACE, 'Reference')

    XMLATTR_Url = ('url', uri.URI.from_octets, str)

    def __init__(self, parent):
        EDMXElement.__init__(self, parent)
        self.url = None


class AnnotationsReference(EDMXElement):
    XMLNAME = (EDMX_NAMESPACE, 'AnnotationsReference')

    XMLATTR_Url = ('url', uri.URI.from_octets, str)

    def __init__(self, parent):
        EDMXElement.__init__(self, parent)
        self.url = None


class Edmx(EDMXElement):

    """Represents the Edmx root element."""
    XMLNAME = (EDMX_NAMESPACE, 'Edmx')
    DataServicesClass = DataServices

    XMLATTR_Version = 'version'

    def __init__(self, parent):
        EDMXElement.__init__(self, parent)
        self.version = "1.0"
        self.Reference = []
        self.AnnotationsReference = []
        self.DataServices = self.DataServicesClass(self)

    def get_children(self):
        for child in itertools.chain(
                self.Reference,
                self.AnnotationsReference):
            yield child
        yield self.DataServices
        for child in EDMXElement.get_children(self):
            yield child

    def validate(self):
        self.DataServices.validate()


class Document(xmlns.XMLNSDocument):

    """Represents an Edmx document."""

    classMap = {}

    def __init__(self, **args):
        xmlns.XMLNSDocument.__init__(self, **args)
        self.defaultNS = EDMX_NAMESPACE
        self.make_prefix(EDMX_NAMESPACE, 'edmx')

    @classmethod
    def get_element_class(cls, name):
        """Overridden to look up name in the class map"""
        eclass = Document.classMap.get(
            name, Document.classMap.get((name[0], None), xmlns.XMLNSElement))
        return eclass

    def validate(self):
        # These extensions MUST be used by a data service in conjunction
        # with the "dataservices" node
        if not isinstance(self.root.DataServices, DataServices):
            raise edm.InvalidMetadataDocument("Expected dataservices node")
        self.root.validate()


xmlns.map_class_elements(Document.classMap, globals())
xmlns.map_class_elements(Document.classMap, edm, edm.NAMESPACE_ALIASES)



""" Python Character Mapping Codec symbol generated from 'codecs/SYMBOL.TXT'
with gencodec.py.

"""  # "

import codecs

from .py2 import range3


# Codec APIs


class Codec(codecs.Codec):

    def encode(self, input, errors='strict'):
        return codecs.charmap_encode(input, errors, encoding_map)

    def decode(self, input, errors='strict'):
        return codecs.charmap_decode(input, errors, decoding_map)


class IncrementalEncoder(codecs.IncrementalEncoder):

    def encode(self, input, final=False):
        return codecs.charmap_encode(input, self.errors, encoding_map)[0]


class IncrementalDecoder(codecs.IncrementalDecoder):

    def decode(self, input, final=False):
        return codecs.charmap_decode(input, self.errors, decoding_map)[0]


class StreamWriter(Codec, codecs.StreamWriter):
    pass


class StreamReader(Codec, codecs.StreamReader):
    pass

# encodings module API


def getregentry():
    return codecs.CodecInfo(
        name='apple-symbol',
        encode=Codec().encode,
        decode=Codec().decode,
        incrementalencoder=IncrementalEncoder,
        incrementaldecoder=IncrementalDecoder,
        streamreader=StreamReader,
        streamwriter=StreamWriter,
    )


# Decoding Map

decoding_map = codecs.make_identity_dict(list(range3(256)))
decoding_map.update({
    0x22: 0x2200,  # FOR ALL
    0x24: 0x2203,  # THERE EXISTS
    0x27: 0x220D,  # SMALL CONTAINS AS MEMBER
    0x2A: 0x2217,  # ASTERISK OPERATOR
    0x2D: 0x2212,  # MINUS SIGN
    0x40: 0x2245,  # APPROXIMATELY EQUAL TO
    0x41: 0x0391,  # GREEK CAPITAL LETTER ALPHA
    0x42: 0x0392,  # GREEK CAPITAL LETTER BETA
    0x43: 0x03A7,  # GREEK CAPITAL LETTER CHI
    0x44: 0x0394,  # GREEK CAPITAL LETTER DELTA
    0x45: 0x0395,  # GREEK CAPITAL LETTER EPSILON
    0x46: 0x03A6,  # GREEK CAPITAL LETTER PHI
    0x47: 0x0393,  # GREEK CAPITAL LETTER GAMMA
    0x48: 0x0397,  # GREEK CAPITAL LETTER ETA
    0x49: 0x0399,  # GREEK CAPITAL LETTER IOTA
    0x4A: 0x03D1,  # GREEK THETA SYMBOL
    0x4B: 0x039A,  # GREEK CAPITAL LETTER KAPPA
    0x4C: 0x039B,  # GREEK CAPITAL LETTER LAMDA
    0x4D: 0x039C,  # GREEK CAPITAL LETTER MU
    0x4E: 0x039D,  # GREEK CAPITAL LETTER NU
    0x4F: 0x039F,  # GREEK CAPITAL LETTER OMICRON
    0x50: 0x03A0,  # GREEK CAPITAL LETTER PI
    0x51: 0x0398,  # GREEK CAPITAL LETTER THETA
    0x52: 0x03A1,  # GREEK CAPITAL LETTER RHO
    0x53: 0x03A3,  # GREEK CAPITAL LETTER SIGMA
    0x54: 0x03A4,  # GREEK CAPITAL LETTER TAU
    0x55: 0x03A5,  # GREEK CAPITAL LETTER UPSILON
    0x56: 0x03C2,  # GREEK SMALL LETTER FINAL SIGMA
    0x57: 0x03A9,  # GREEK CAPITAL LETTER OMEGA
    0x58: 0x039E,  # GREEK CAPITAL LETTER XI
    0x59: 0x03A8,  # GREEK CAPITAL LETTER PSI
    0x5A: 0x0396,  # GREEK CAPITAL LETTER ZETA
    0x5C: 0x2234,  # THEREFORE
    0x5E: 0x22A5,  # UP TACK
    0x60: 0xF8E5,  # radical extender # corporate char
    0x61: 0x03B1,  # GREEK SMALL LETTER ALPHA
    0x62: 0x03B2,  # GREEK SMALL LETTER BETA
    0x63: 0x03C7,  # GREEK SMALL LETTER CHI
    0x64: 0x03B4,  # GREEK SMALL LETTER DELTA
    0x65: 0x03B5,  # GREEK SMALL LETTER EPSILON
    0x66: 0x03C6,  # GREEK SMALL LETTER PHI
    0x67: 0x03B3,  # GREEK SMALL LETTER GAMMA
    0x68: 0x03B7,  # GREEK SMALL LETTER ETA
    0x69: 0x03B9,  # GREEK SMALL LETTER IOTA
    0x6A: 0x03D5,  # GREEK PHI SYMBOL
    0x6B: 0x03BA,  # GREEK SMALL LETTER KAPPA
    0x6C: 0x03BB,  # GREEK SMALL LETTER LAMDA
    0x6D: 0x03BC,  # GREEK SMALL LETTER MU
    0x6E: 0x03BD,  # GREEK SMALL LETTER NU
    0x6F: 0x03BF,  # GREEK SMALL LETTER OMICRON
    0x70: 0x03C0,  # GREEK SMALL LETTER PI
    0x71: 0x03B8,  # GREEK SMALL LETTER THETA
    0x72: 0x03C1,  # GREEK SMALL LETTER RHO
    0x73: 0x03C3,  # GREEK SMALL LETTER SIGMA
    0x74: 0x03C4,  # GREEK SMALL LETTER TAU
    0x75: 0x03C5,  # GREEK SMALL LETTER UPSILON
    0x76: 0x03D6,  # GREEK PI SYMBOL
    0x77: 0x03C9,  # GREEK SMALL LETTER OMEGA
    0x78: 0x03BE,  # GREEK SMALL LETTER XI
    0x79: 0x03C8,  # GREEK SMALL LETTER PSI
    0x7A: 0x03B6,  # GREEK SMALL LETTER ZETA
    0x7E: 0x223C,  # TILDE OPERATOR
    0x80: None,
    0x81: None,
    0x82: None,
    0x83: None,
    0x84: None,
    0x85: None,
    0x86: None,
    0x87: None,
    0x88: None,
    0x89: None,
    0x8A: None,
    0x8B: None,
    0x8C: None,
    0x8D: None,
    0x8E: None,
    0x8F: None,
    0x90: None,
    0x91: None,
    0x92: None,
    0x93: None,
    0x94: None,
    0x95: None,
    0x96: None,
    0x97: None,
    0x98: None,
    0x99: None,
    0x9A: None,
    0x9B: None,
    0x9C: None,
    0x9D: None,
    0x9E: None,
    0x9F: None,
    0xA0: 0x20AC,  # EURO SIGN
    0xA1: 0x03D2,  # GREEK UPSILON WITH HOOK SYMBOL
    0xA2: 0x2032,  # PRIME # minute
    0xA3: 0x2264,  # LESS-THAN OR EQUAL TO
    0xA4: 0x2044,  # FRACTION SLASH
    0xA5: 0x221E,  # INFINITY
    0xA6: 0x0192,  # LATIN SMALL LETTER F WITH HOOK
    0xA7: 0x2663,  # BLACK CLUB SUIT
    0xA8: 0x2666,  # BLACK DIAMOND SUIT
    0xA9: 0x2665,  # BLACK HEART SUIT
    0xAA: 0x2660,  # BLACK SPADE SUIT
    0xAB: 0x2194,  # LEFT RIGHT ARROW
    0xAC: 0x2190,  # LEFTWARDS ARROW
    0xAD: 0x2191,  # UPWARDS ARROW
    0xAE: 0x2192,  # RIGHTWARDS ARROW
    0xAF: 0x2193,  # DOWNWARDS ARROW
    0xB2: 0x2033,  # DOUBLE PRIME # second
    0xB3: 0x2265,  # GREATER-THAN OR EQUAL TO
    0xB4: 0x00D7,  # MULTIPLICATION SIGN
    0xB5: 0x221D,  # PROPORTIONAL TO
    0xB6: 0x2202,  # PARTIAL DIFFERENTIAL
    0xB7: 0x2022,  # BULLET
    0xB8: 0x00F7,  # DIVISION SIGN
    0xB9: 0x2260,  # NOT EQUAL TO
    0xBA: 0x2261,  # IDENTICAL TO
    0xBB: 0x2248,  # ALMOST EQUAL TO
    0xBC: 0x2026,  # HORIZONTAL ELLIPSIS
    # VERTICAL LINE EXTENSION (for arrows) # for Unicode 4.0 and later
    0xBD: 0x23D0,
    # HORIZONTAL LINE EXTENSION (for arrows) # for Unicode 3.2 and later
    0xBE: 0x23AF,
    0xBF: 0x21B5,  # DOWNWARDS ARROW WITH CORNER LEFTWARDS
    0xC0: 0x2135,  # ALEF SYMBOL
    0xC1: 0x2111,  # BLACK-LETTER CAPITAL I
    0xC2: 0x211C,  # BLACK-LETTER CAPITAL R
    0xC3: 0x2118,  # SCRIPT CAPITAL P
    0xC4: 0x2297,  # CIRCLED TIMES
    0xC5: 0x2295,  # CIRCLED PLUS
    0xC6: 0x2205,  # EMPTY SET
    0xC7: 0x2229,  # INTERSECTION
    0xC8: 0x222A,  # UNION
    0xC9: 0x2283,  # SUPERSET OF
    0xCA: 0x2287,  # SUPERSET OF OR EQUAL TO
    0xCB: 0x2284,  # NOT A SUBSET OF
    0xCC: 0x2282,  # SUBSET OF
    0xCD: 0x2286,  # SUBSET OF OR EQUAL TO
    0xCE: 0x2208,  # ELEMENT OF
    0xCF: 0x2209,  # NOT AN ELEMENT OF
    0xD0: 0x2220,  # ANGLE
    0xD1: 0x2207,  # NABLA
    0xD2: 0x00AE,  # REGISTERED SIGN # serif
    0xD3: 0x00A9,  # COPYRIGHT SIGN # serif
    0xD4: 0x2122,  # TRADE MARK SIGN # serif
    0xD5: 0x220F,  # N-ARY PRODUCT
    0xD6: 0x221A,  # SQUARE ROOT
    0xD7: 0x22C5,  # DOT OPERATOR
    0xD8: 0x00AC,  # NOT SIGN
    0xD9: 0x2227,  # LOGICAL AND
    0xDA: 0x2228,  # LOGICAL OR
    0xDB: 0x21D4,  # LEFT RIGHT DOUBLE ARROW
    0xDC: 0x21D0,  # LEFTWARDS DOUBLE ARROW
    0xDD: 0x21D1,  # UPWARDS DOUBLE ARROW
    0xDE: 0x21D2,  # RIGHTWARDS DOUBLE ARROW
    0xDF: 0x21D3,  # DOWNWARDS DOUBLE ARROW
    0xE0: 0x25CA,  # LOZENGE # previously mapped to 0x22C4 DIAMOND OPERATOR
    0xE1: 0x3008,  # LEFT ANGLE BRACKET
    0xE2: (0x00AE, 0xF87F),  # REGISTERED SIGN, alternate: sans serif
    0xE3: (0x00A9, 0xF87F),  # COPYRIGHT SIGN, alternate: sans serif
    0xE4: (0x2122, 0xF87F),  # TRADE MARK SIGN, alternate: sans serif
    0xE5: 0x2211,  # N-ARY SUMMATION
    0xE6: 0x239B,  # LEFT PARENTHESIS UPPER HOOK # for Unicode 3.2 and later
    0xE7: 0x239C,  # LEFT PARENTHESIS EXTENSION # for Unicode 3.2 and later
    0xE8: 0x239D,  # LEFT PARENTHESIS LOWER HOOK # for Unicode 3.2 and later
    # LEFT SQUARE BRACKET UPPER CORNER # for Unicode 3.2 and later
    0xE9: 0x23A1,
    0xEA: 0x23A2,  # LEFT SQUARE BRACKET EXTENSION # for Unicode 3.2 and later
    # LEFT SQUARE BRACKET LOWER CORNER # for Unicode 3.2 and later
    0xEB: 0x23A3,
    0xEC: 0x23A7,  # LEFT CURLY BRACKET UPPER HOOK # for Unicode 3.2 and later
    # LEFT CURLY BRACKET MIDDLE PIECE # for Unicode 3.2 and later
    0xED: 0x23A8,
    0xEE: 0x23A9,  # LEFT CURLY BRACKET LOWER HOOK # for Unicode 3.2 and later
    0xEF: 0x23AA,  # CURLY BRACKET EXTENSION # for Unicode 3.2 and later
    0xF0: 0xF8FF,  # Apple logo
    0xF1: 0x3009,  # RIGHT ANGLE BRACKET
    0xF2: 0x222B,  # INTEGRAL
    0xF3: 0x2320,  # TOP HALF INTEGRAL
    0xF4: 0x23AE,  # INTEGRAL EXTENSION # for Unicode 3.2 and later
    0xF5: 0x2321,  # BOTTOM HALF INTEGRAL
    0xF6: 0x239E,  # RIGHT PARENTHESIS UPPER HOOK # for Unicode 3.2 and later
    0xF7: 0x239F,  # RIGHT PARENTHESIS EXTENSION # for Unicode 3.2 and later
    0xF8: 0x23A0,  # RIGHT PARENTHESIS LOWER HOOK # for Unicode 3.2 and later
    # RIGHT SQUARE BRACKET UPPER CORNER # for Unicode 3.2 and later
    0xF9: 0x23A4,
    0xFA: 0x23A5,  # RIGHT SQUARE BRACKET EXTENSION # for Unicode 3.2 and later
    # RIGHT SQUARE BRACKET LOWER CORNER # for Unicode 3.2 and later
    0xFB: 0x23A6,
    0xFC: 0x23AB,  # RIGHT CURLY BRACKET UPPER HOOK # for Unicode 3.2 and later
    # RIGHT CURLY BRACKET MIDDLE PIECE # for Unicode 3.2 and later
    0xFD: 0x23AC,
    0xFE: 0x23AD,  # RIGHT CURLY BRACKET LOWER HOOK # for Unicode 3.2 and later
    0xFF: None,
})

# Encoding Map

encoding_map = {
    0x0000: 0x00,  # CONTROL CHARACTER
    0x0001: 0x01,  # CONTROL CHARACTER
    0x0002: 0x02,  # CONTROL CHARACTER
    0x0003: 0x03,  # CONTROL CHARACTER
    0x0004: 0x04,  # CONTROL CHARACTER
    0x0005: 0x05,  # CONTROL CHARACTER
    0x0006: 0x06,  # CONTROL CHARACTER
    0x0007: 0x07,  # CONTROL CHARACTER
    0x0008: 0x08,  # CONTROL CHARACTER
    0x0009: 0x09,  # CONTROL CHARACTER
    0x000A: 0x0A,  # CONTROL CHARACTER
    0x000B: 0x0B,  # CONTROL CHARACTER
    0x000C: 0x0C,  # CONTROL CHARACTER
    0x000D: 0x0D,  # CONTROL CHARACTER
    0x000E: 0x0E,  # CONTROL CHARACTER
    0x000F: 0x0F,  # CONTROL CHARACTER
    0x0010: 0x10,  # CONTROL CHARACTER
    0x0011: 0x11,  # CONTROL CHARACTER
    0x0012: 0x12,  # CONTROL CHARACTER
    0x0013: 0x13,  # CONTROL CHARACTER
    0x0014: 0x14,  # CONTROL CHARACTER
    0x0015: 0x15,  # CONTROL CHARACTER
    0x0016: 0x16,  # CONTROL CHARACTER
    0x0017: 0x17,  # CONTROL CHARACTER
    0x0018: 0x18,  # CONTROL CHARACTER
    0x0019: 0x19,  # CONTROL CHARACTER
    0x001A: 0x1A,  # CONTROL CHARACTER
    0x001B: 0x1B,  # CONTROL CHARACTER
    0x001C: 0x1C,  # CONTROL CHARACTER
    0x001D: 0x1D,  # CONTROL CHARACTER
    0x001E: 0x1E,  # CONTROL CHARACTER
    0x001F: 0x1F,  # CONTROL CHARACTER
    0x0020: 0x20,  # SPACE
    0x0021: 0x21,  # EXCLAMATION MARK
    0x0023: 0x23,  # NUMBER SIGN
    0x0025: 0x25,  # PERCENT SIGN
    0x0026: 0x26,  # AMPERSAND
    0x0028: 0x28,  # LEFT PARENTHESIS
    0x0029: 0x29,  # RIGHT PARENTHESIS
    0x002B: 0x2B,  # PLUS SIGN
    0x002C: 0x2C,  # COMMA
    0x002E: 0x2E,  # FULL STOP
    0x002F: 0x2F,  # SOLIDUS
    0x0030: 0x30,  # DIGIT ZERO
    0x0031: 0x31,  # DIGIT ONE
    0x0032: 0x32,  # DIGIT TWO
    0x0033: 0x33,  # DIGIT THREE
    0x0034: 0x34,  # DIGIT FOUR
    0x0035: 0x35,  # DIGIT FIVE
    0x0036: 0x36,  # DIGIT SIX
    0x0037: 0x37,  # DIGIT SEVEN
    0x0038: 0x38,  # DIGIT EIGHT
    0x0039: 0x39,  # DIGIT NINE
    0x003A: 0x3A,  # COLON
    0x003B: 0x3B,  # SEMICOLON
    0x003C: 0x3C,  # LESS-THAN SIGN
    0x003D: 0x3D,  # EQUALS SIGN
    0x003E: 0x3E,  # GREATER-THAN SIGN
    0x003F: 0x3F,  # QUESTION MARK
    0x005B: 0x5B,  # LEFT SQUARE BRACKET
    0x005D: 0x5D,  # RIGHT SQUARE BRACKET
    0x005F: 0x5F,  # LOW LINE
    0x007B: 0x7B,  # LEFT CURLY BRACKET
    0x007C: 0x7C,  # VERTICAL LINE
    0x007D: 0x7D,  # RIGHT CURLY BRACKET
    0x007F: 0x7F,  # CONTROL CHARACTER
    0x00A9: 0xD3,  # COPYRIGHT SIGN # serif
    0x00AC: 0xD8,  # NOT SIGN
    0x00AE: 0xD2,  # REGISTERED SIGN # serif
    0x00B0: 0xB0,  # DEGREE SIGN
    0x00B1: 0xB1,  # PLUS-MINUS SIGN
    0x00D7: 0xB4,  # MULTIPLICATION SIGN
    0x00F7: 0xB8,  # DIVISION SIGN
    0x0192: 0xA6,  # LATIN SMALL LETTER F WITH HOOK
    0x0391: 0x41,  # GREEK CAPITAL LETTER ALPHA
    0x0392: 0x42,  # GREEK CAPITAL LETTER BETA
    0x0393: 0x47,  # GREEK CAPITAL LETTER GAMMA
    0x0394: 0x44,  # GREEK CAPITAL LETTER DELTA
    0x0395: 0x45,  # GREEK CAPITAL LETTER EPSILON
    0x0396: 0x5A,  # GREEK CAPITAL LETTER ZETA
    0x0397: 0x48,  # GREEK CAPITAL LETTER ETA
    0x0398: 0x51,  # GREEK CAPITAL LETTER THETA
    0x0399: 0x49,  # GREEK CAPITAL LETTER IOTA
    0x039A: 0x4B,  # GREEK CAPITAL LETTER KAPPA
    0x039B: 0x4C,  # GREEK CAPITAL LETTER LAMDA
    0x039C: 0x4D,  # GREEK CAPITAL LETTER MU
    0x039D: 0x4E,  # GREEK CAPITAL LETTER NU
    0x039E: 0x58,  # GREEK CAPITAL LETTER XI
    0x039F: 0x4F,  # GREEK CAPITAL LETTER OMICRON
    0x03A0: 0x50,  # GREEK CAPITAL LETTER PI
    0x03A1: 0x52,  # GREEK CAPITAL LETTER RHO
    0x03A3: 0x53,  # GREEK CAPITAL LETTER SIGMA
    0x03A4: 0x54,  # GREEK CAPITAL LETTER TAU
    0x03A5: 0x55,  # GREEK CAPITAL LETTER UPSILON
    0x03A6: 0x46,  # GREEK CAPITAL LETTER PHI
    0x03A7: 0x43,  # GREEK CAPITAL LETTER CHI
    0x03A8: 0x59,  # GREEK CAPITAL LETTER PSI
    0x03A9: 0x57,  # GREEK CAPITAL LETTER OMEGA
    0x03B1: 0x61,  # GREEK SMALL LETTER ALPHA
    0x03B2: 0x62,  # GREEK SMALL LETTER BETA
    0x03B3: 0x67,  # GREEK SMALL LETTER GAMMA
    0x03B4: 0x64,  # GREEK SMALL LETTER DELTA
    0x03B5: 0x65,  # GREEK SMALL LETTER EPSILON
    0x03B6: 0x7A,  # GREEK SMALL LETTER ZETA
    0x03B7: 0x68,  # GREEK SMALL LETTER ETA
    0x03B8: 0x71,  # GREEK SMALL LETTER THETA
    0x03B9: 0x69,  # GREEK SMALL LETTER IOTA
    0x03BA: 0x6B,  # GREEK SMALL LETTER KAPPA
    0x03BB: 0x6C,  # GREEK SMALL LETTER LAMDA
    0x03BC: 0x6D,  # GREEK SMALL LETTER MU
    0x03BD: 0x6E,  # GREEK SMALL LETTER NU
    0x03BE: 0x78,  # GREEK SMALL LETTER XI
    0x03BF: 0x6F,  # GREEK SMALL LETTER OMICRON
    0x03C0: 0x70,  # GREEK SMALL LETTER PI
    0x03C1: 0x72,  # GREEK SMALL LETTER RHO
    0x03C2: 0x56,  # GREEK SMALL LETTER FINAL SIGMA
    0x03C3: 0x73,  # GREEK SMALL LETTER SIGMA
    0x03C4: 0x74,  # GREEK SMALL LETTER TAU
    0x03C5: 0x75,  # GREEK SMALL LETTER UPSILON
    0x03C6: 0x66,  # GREEK SMALL LETTER PHI
    0x03C7: 0x63,  # GREEK SMALL LETTER CHI
    0x03C8: 0x79,  # GREEK SMALL LETTER PSI
    0x03C9: 0x77,  # GREEK SMALL LETTER OMEGA
    0x03D1: 0x4A,  # GREEK THETA SYMBOL
    0x03D2: 0xA1,  # GREEK UPSILON WITH HOOK SYMBOL
    0x03D5: 0x6A,  # GREEK PHI SYMBOL
    0x03D6: 0x76,  # GREEK PI SYMBOL
    0x2022: 0xB7,  # BULLET
    0x2026: 0xBC,  # HORIZONTAL ELLIPSIS
    0x2032: 0xA2,  # PRIME # minute
    0x2033: 0xB2,  # DOUBLE PRIME # second
    0x2044: 0xA4,  # FRACTION SLASH
    0x20AC: 0xA0,  # EURO SIGN
    0x2111: 0xC1,  # BLACK-LETTER CAPITAL I
    0x2118: 0xC3,  # SCRIPT CAPITAL P
    0x211C: 0xC2,  # BLACK-LETTER CAPITAL R
    0x2122: 0xD4,  # TRADE MARK SIGN # serif
    0x2135: 0xC0,  # ALEF SYMBOL
    0x2190: 0xAC,  # LEFTWARDS ARROW
    0x2191: 0xAD,  # UPWARDS ARROW
    0x2192: 0xAE,  # RIGHTWARDS ARROW
    0x2193: 0xAF,  # DOWNWARDS ARROW
    0x2194: 0xAB,  # LEFT RIGHT ARROW
    0x21B5: 0xBF,  # DOWNWARDS ARROW WITH CORNER LEFTWARDS
    0x21D0: 0xDC,  # LEFTWARDS DOUBLE ARROW
    0x21D1: 0xDD,  # UPWARDS DOUBLE ARROW
    0x21D2: 0xDE,  # RIGHTWARDS DOUBLE ARROW
    0x21D3: 0xDF,  # DOWNWARDS DOUBLE ARROW
    0x21D4: 0xDB,  # LEFT RIGHT DOUBLE ARROW
    0x2200: 0x22,  # FOR ALL
    0x2202: 0xB6,  # PARTIAL DIFFERENTIAL
    0x2203: 0x24,  # THERE EXISTS
    0x2205: 0xC6,  # EMPTY SET
    0x2207: 0xD1,  # NABLA
    0x2208: 0xCE,  # ELEMENT OF
    0x2209: 0xCF,  # NOT AN ELEMENT OF
    0x220D: 0x27,  # SMALL CONTAINS AS MEMBER
    0x220F: 0xD5,  # N-ARY PRODUCT
    0x2211: 0xE5,  # N-ARY SUMMATION
    0x2212: 0x2D,  # MINUS SIGN
    0x2217: 0x2A,  # ASTERISK OPERATOR
    0x221A: 0xD6,  # SQUARE ROOT
    0x221D: 0xB5,  # PROPORTIONAL TO
    0x221E: 0xA5,  # INFINITY
    0x2220: 0xD0,  # ANGLE
    0x2227: 0xD9,  # LOGICAL AND
    0x2228: 0xDA,  # LOGICAL OR
    0x2229: 0xC7,  # INTERSECTION
    0x222A: 0xC8,  # UNION
    0x222B: 0xF2,  # INTEGRAL
    0x2234: 0x5C,  # THEREFORE
    0x223C: 0x7E,  # TILDE OPERATOR
    0x2245: 0x40,  # APPROXIMATELY EQUAL TO
    0x2248: 0xBB,  # ALMOST EQUAL TO
    0x2260: 0xB9,  # NOT EQUAL TO
    0x2261: 0xBA,  # IDENTICAL TO
    0x2264: 0xA3,  # LESS-THAN OR EQUAL TO
    0x2265: 0xB3,  # GREATER-THAN OR EQUAL TO
    0x2282: 0xCC,  # SUBSET OF
    0x2283: 0xC9,  # SUPERSET OF
    0x2284: 0xCB,  # NOT A SUBSET OF
    0x2286: 0xCD,  # SUBSET OF OR EQUAL TO
    0x2287: 0xCA,  # SUPERSET OF OR EQUAL TO
    0x2295: 0xC5,  # CIRCLED PLUS
    0x2297: 0xC4,  # CIRCLED TIMES
    0x22A5: 0x5E,  # UP TACK
    0x22C5: 0xD7,  # DOT OPERATOR
    0x2320: 0xF3,  # TOP HALF INTEGRAL
    0x2321: 0xF5,  # BOTTOM HALF INTEGRAL
    0x239B: 0xE6,  # LEFT PARENTHESIS UPPER HOOK # for Unicode 3.2 and later
    0x239C: 0xE7,  # LEFT PARENTHESIS EXTENSION # for Unicode 3.2 and later
    0x239D: 0xE8,  # LEFT PARENTHESIS LOWER HOOK # for Unicode 3.2 and later
    0x239E: 0xF6,  # RIGHT PARENTHESIS UPPER HOOK # for Unicode 3.2 and later
    0x239F: 0xF7,  # RIGHT PARENTHESIS EXTENSION # for Unicode 3.2 and later
    0x23A0: 0xF8,  # RIGHT PARENTHESIS LOWER HOOK # for Unicode 3.2 and later
    # LEFT SQUARE BRACKET UPPER CORNER # for Unicode 3.2 and later
    0x23A1: 0xE9,
    0x23A2: 0xEA,  # LEFT SQUARE BRACKET EXTENSION # for Unicode 3.2 and later
    # LEFT SQUARE BRACKET LOWER CORNER # for Unicode 3.2 and later
    0x23A3: 0xEB,
    # RIGHT SQUARE BRACKET UPPER CORNER # for Unicode 3.2 and later
    0x23A4: 0xF9,
    0x23A5: 0xFA,  # RIGHT SQUARE BRACKET EXTENSION # for Unicode 3.2 and later
    # RIGHT SQUARE BRACKET LOWER CORNER # for Unicode 3.2 and later
    0x23A6: 0xFB,
    0x23A7: 0xEC,  # LEFT CURLY BRACKET UPPER HOOK # for Unicode 3.2 and later
    # LEFT CURLY BRACKET MIDDLE PIECE # for Unicode 3.2 and later
    0x23A8: 0xED,
    0x23A9: 0xEE,  # LEFT CURLY BRACKET LOWER HOOK # for Unicode 3.2 and later
    0x23AA: 0xEF,  # CURLY BRACKET EXTENSION # for Unicode 3.2 and later
    0x23AB: 0xFC,  # RIGHT CURLY BRACKET UPPER HOOK # for Unicode 3.2 and later
    # RIGHT CURLY BRACKET MIDDLE PIECE # for Unicode 3.2 and later
    0x23AC: 0xFD,
    0x23AD: 0xFE,  # RIGHT CURLY BRACKET LOWER HOOK # for Unicode 3.2 and later
    0x23AE: 0xF4,  # INTEGRAL EXTENSION # for Unicode 3.2 and later
    # HORIZONTAL LINE EXTENSION (for arrows) # for Unicode 3.2 and later
    0x23AF: 0xBE,
    # VERTICAL LINE EXTENSION (for arrows) # for Unicode 4.0 and later
    0x23D0: 0xBD,
    0x25CA: 0xE0,  # LOZENGE # previously mapped to 0x22C4 DIAMOND OPERATOR
    0x2660: 0xAA,  # BLACK SPADE SUIT
    0x2663: 0xA7,  # BLACK CLUB SUIT
    0x2665: 0xA9,  # BLACK HEART SUIT
    0x2666: 0xA8,  # BLACK DIAMOND SUIT
    0x3008: 0xE1,  # LEFT ANGLE BRACKET
    0x3009: 0xF1,  # RIGHT ANGLE BRACKET
    0xF8E5: 0x60,  # radical extender # corporate char
    0xF8FF: 0xF0,  # Apple logo
    (0x00A9, 0xF87F): 0xE3,  # COPYRIGHT SIGN, alternate: sans serif
    (0x00AE, 0xF87F): 0xE2,  # REGISTERED SIGN, alternate: sans serif
    (0x2122, 0xF87F): 0xE4,  # TRADE MARK SIGN, alternate: sans serif
}



#! /usr/bin/env python
"""This module implements the Basic LTI 1.0 specification defined by IMS GLC

This is an experimental module and subject to significant change in future
releases.  Use with caution.
"""

import logging
import os
import random
import time

from hashlib import sha256

from pyslet import iso8601 as iso
from pyslet import wsgi
from pyslet.odata2 import csdl as edm
from pyslet.odata2 import core as odata
from pyslet.odata2 import metadata as edmx
from pyslet.odata2 import sqlds as sql
from pyslet.py2 import (
    byte_value,
    dict_items,
    dict_keys,
    force_bytes,
    long2,
    parse_qs,
    range3,
    ul
    )
from pyslet.pep8 import MigratedClass, old_method
from pyslet.rfc2396 import URI
from pyslet.urn import URN
from pyslet.xml import structures as xml


try:
    from oauthlib import oauth1 as oauth
    got_oauth = True
except ImportError:
    # at some point in the future, we may just replace the signature
    # verification code with something from the newer oauthlib, if
    # similar functionality can easily be exposed.
    got_oauth = False

    class OAuthMissing(object):

        def __init__(*args, **kwargs):
            raise RuntimeError("oauthlib module required")

    class oauth(object):    # noqa
        RequestValidator = OAuthMissing
        SignatureOnlyEndpoint = OAuthMissing


#: The version of LTI we support
LTI_VERSION = "LTI-1p0"

#: The message type we support
LTI_MESSAGE_TYPE = "basic-lti-launch-request"


class LTIError(Exception):

    """Base class for all LTI errors"""
    pass


class LTIAuthenticationError(LTIError):

    """Indicates an authentication error (on launch)"""
    pass


class LTIProtocolError(LTIError):

    """Indicates a protocol violoation

    This may be raised if the message type or protocol version in a
    launch request do not match the expected values or if a required
    parameter is missing."""
    pass


#: A mapping from a system role handle to the full URN for the role as a
#: :class:`~pyslet.rfc2396.URI` instance.
SYSROLE_HANDLES = {
    'SysAdmin': URI.from_octets('urn:lti:sysrole:ims/lis/SysAdmin'),
    'SysSupport': URI.from_octets('urn:lti:sysrole:ims/lis/SysSupport'),
    'Creator': URI.from_octets('urn:lti:sysrole:ims/lis/Creator'),
    'AccountAdmin': URI.from_octets('urn:lti:sysrole:ims/lis/AccountAdmin'),
    'User': URI.from_octets('urn:lti:sysrole:ims/lis/User'),
    'Administrator': URI.from_octets('urn:lti:sysrole:ims/lis/Administrator'),
    'None': URI.from_octets('urn:lti:sysrole:ims/lis/None')
}

#: A mapping from a institution role handle to the full URN for the role
#: as a :class:`~pyslet.rfc2396.URI` instance.
INSTROLE_HANDLES = {
    'Student': URI.from_octets('urn:lti:instrole:ims/lis/Student'),
    'Faculty': URI.from_octets('urn:lti:instrole:ims/lis/Faculty'),
    'Member': URI.from_octets('urn:lti:instrole:ims/lis/Member'),
    'Learner': URI.from_octets('urn:lti:instrole:ims/lis/Learner'),
    'Instructor': URI.from_octets('urn:lti:instrole:ims/lis/Instructor'),
    'Mentor': URI.from_octets('urn:lti:instrole:ims/lis/Mentor'),
    'Staff': URI.from_octets('urn:lti:instrole:ims/lis/Staff'),
    'Alumni': URI.from_octets('urn:lti:instrole:ims/lis/Alumni'),
    'ProspectiveStudent':
        URI.from_octets('urn:lti:instrole:ims/lis/ProspectiveStudent'),
    'Guest': URI.from_octets('urn:lti:instrole:ims/lis/Guest'),
    'Other': URI.from_octets('urn:lti:instrole:ims/lis/Other'),
    'Administrator': URI.from_octets('urn:lti:instrole:ims/lis/Administrator'),
    'Observer': URI.from_octets('urn:lti:instrole:ims/lis/Observer'),
    'None': URI.from_octets('urn:lti:instrole:ims/lis/None')
}

#: A mapping from LTI role handles to the full URN for the role as a
#: :class:`~pyslet.rfc2396.URI` instance.
ROLE_HANDLES = {
    'Learner': URI.from_octets('urn:lti:role:ims/lis/Learner'),
    'Learner/Learner': URI.from_octets('urn:lti:role:ims/lis/Learner/Learner'),
    'Learner/NonCreditLearner':
        URI.from_octets('urn:lti:role:ims/lis/Learner/NonCreditLearner'),
    'Learner/GuestLearner':
        URI.from_octets('urn:lti:role:ims/lis/Learner/GuestLearner'),
    'Learner/ExternalLearner':
        URI.from_octets('urn:lti:role:ims/lis/Learner/ExternalLearner'),
    'Learner/Instructor':
        URI.from_octets('urn:lti:role:ims/lis/Learner/Instructor'),
    'Instructor': URI.from_octets('urn:lti:role:ims/lis/Instructor'),
    'Instructor/PrimaryInstructor':
        URI.from_octets('urn:lti:role:ims/lis/Instructor/PrimaryInstructor'),
    'Instructor/Lecturer':
        URI.from_octets('urn:lti:role:ims/lis/Instructor/Lecturer'),
    'Instructor/GuestInstructor':
        URI.from_octets('urn:lti:role:ims/lis/Instructor/GuestInstructor'),
    'Instructor/ExternalInstructor':
        URI.from_octets('urn:lti:role:ims/lis/Instructor/ExternalInstructor'),
    'ContentDeveloper':
        URI.from_octets('urn:lti:role:ims/lis/ContentDeveloper'),
    'ContentDeveloper/ContentDeveloper':
        URI.from_octets('urn:lti:role:ims/lis/ContentDeveloper/'
                        'ContentDeveloper'),
    'ContentDeveloper/Librarian':
        URI.from_octets('urn:lti:role:ims/lis/ContentDeveloper/Librarian'),
    'ContentDeveloper/ContentExpert':
        URI.from_octets('urn:lti:role:ims/lis/ContentDeveloper/ContentExpert'),
    'ContentDeveloper/ExternalContentExpert':
        URI.from_octets('urn:lti:role:ims/lis/ContentDeveloper/'
                        'ExternalContentExpert'),
    'Member': URI.from_octets('urn:lti:role:ims/lis/Member'),
    'Member/Member': URI.from_octets('urn:lti:role:ims/lis/Member/Member'),
    'Manager': URI.from_octets('urn:lti:role:ims/lis/Manager'),
    'Manager/AreaManager':
        URI.from_octets('urn:lti:role:ims/lis/Manager/AreaManager'),
    'Manager/CourseCoordinator':
        URI.from_octets('urn:lti:role:ims/lis/Manager/CourseCoordinator'),
    'Manager/Observer':
        URI.from_octets('urn:lti:role:ims/lis/Manager/Observer'),
    'Manager/ExternalObserver':
        URI.from_octets('urn:lti:role:ims/lis/Manager/ExternalObserver'),
    'Mentor': URI.from_octets('urn:lti:role:ims/lis/Mentor'),
    'Mentor/Mentor': URI.from_octets('urn:lti:role:ims/lis/Mentor/Mentor'),
    'Mentor/Reviewer': URI.from_octets('urn:lti:role:ims/lis/Mentor/Reviewer'),
    'Mentor/Advisor': URI.from_octets('urn:lti:role:ims/lis/Mentor/Advisor'),
    'Mentor/Auditor': URI.from_octets('urn:lti:role:ims/lis/Mentor/Auditor'),
    'Mentor/Tutor': URI.from_octets('urn:lti:role:ims/lis/Mentor/Tutor'),
    'Mentor/LearningFacilitator':
        URI.from_octets('urn:lti:role:ims/lis/Mentor/LearningFacilitator'),
    'Mentor/ExternalMentor':
        URI.from_octets('urn:lti:role:ims/lis/Mentor/ExternalMentor'),
    'Mentor/ExternalReviewer':
        URI.from_octets('urn:lti:role:ims/lis/Mentor/ExternalReviewer'),
    'Mentor/ExternalAdvisor':
        URI.from_octets('urn:lti:role:ims/lis/Mentor/ExternalAdvisor'),
    'Mentor/ExternalAuditor':
        URI.from_octets('urn:lti:role:ims/lis/Mentor/ExternalAuditor'),
    'Mentor/ExternalTutor':
        URI.from_octets('urn:lti:role:ims/lis/Mentor/ExternalTutor'),
    'Mentor/ExternalLearningFacilitator':
        URI.from_octets('urn:lti:role:ims/lis/Mentor/'
                        'ExternalLearningFacilitator'),
    'Administrator': URI.from_octets('urn:lti:role:ims/lis/Administrator'),
    'Administrator/Administrator':
        URI.from_octets('urn:lti:role:ims/lis/Administrator/Administrator'),
    'Administrator/Support':
        URI.from_octets('urn:lti:role:ims/lis/Administrator/Support'),
    'Administrator/Developer':
        URI.from_octets('urn:lti:role:ims/lis/Administrator/'
                        'ExternalDeveloper'),
    'Administrator/SystemAdministrator':
        URI.from_octets('urn:lti:role:ims/lis/Administrator/'
                        'SystemAdministrator'),
    'Administrator/ExternalSystemAdministrator':
        URI.from_octets('urn:lti:role:ims/lis/Administrator/'
                        'ExternalSystemAdministrator'),
    'Administrator/ExternalDeveloper':
        URI.from_octets('urn:lti:role:ims/lis/Administrator/'
                        'ExternalDeveloper'),
    'Administrator/ExternalSupport':
        URI.from_octets('urn:lti:role:ims/lis/Administrator/ExternalSupport'),
    'TeachingAssistant':
        URI.from_octets('urn:lti:role:ims/lis/''TeachingAssistant'),
    'TeachingAssistant/TeachingAssistant':
        URI.from_octets('urn:lti:role:ims/lis/TeachingAssistant/'
                        'TeachingAssistant'),
    'TeachingAssistant/TeachingAssistantSection':
        URI.from_octets('urn:lti:role:ims/lis/TeachingAssistant/'
                        'TeachingAssistantSection'),
    'TeachingAssistant/TeachingAssistantSectionAssociation':
        URI.from_octets('urn:lti:role:ims/lis/TeachingAssistant/'
                        'TeachingAssistantSectionAssociation'),
    'TeachingAssistant/TeachingAssistantOffering':
        URI.from_octets('urn:lti:role:ims/lis/TeachingAssistant/'
                        'TeachingAssistantOffering'),
    'TeachingAssistant/TeachingAssistantTemplate':
        URI.from_octets('urn:lti:role:ims/lis/TeachingAssistant/'
                        'TeachingAssistantTemplate'),
    'TeachingAssistant/TeachingAssistantGroup':
        URI.from_octets('urn:lti:role:ims/lis/TeachingAssistant/'
                        'TeachingAssistantGroup'),
    'TeachingAssistant/Grader':
        URI.from_octets('urn:lti:role:ims/lis/TeachingAssistant/Grader')
}


def split_role(role):
    """Splits an LTI role into vocab, type and sub-type

    role
        A :class:`~pyslet.urn.URN` instance containing the
        full definition of the role.

    Returns a triple of:

    vocab
        One of 'role', 'sysrole', 'instrole' or some future vocab
        extension.

    rtype
        The role type, e.g., 'Learner', 'Instructor'

    rsubtype
        The role sub-type , e.g., 'NonCreditLearner', 'Lecturer'. Will
        be None if there is no sub-type.

    If this is not an LTI defined role, or the role descriptor does not
    start with the path ims/lis then ValueError is raised."""
    if not isinstance(role, URN) or role.nid != 'lti':
        raise ValueError("Expected lti role: %s" % str(role))
    rvocab = ims = lis = rtype = rsubtype = None
    srole = role.nss.split(":")
    if len(srole) == 2:
        rvocab = srole[0]
        srole = srole[1].split('/')
        if len(srole) == 3:
            ims = srole[0]
            lis = srole[1]
            rtype = srole[2]
        elif len(srole) == 4:
            ims = srole[0]
            lis = srole[1]
            rtype = srole[2]
            rsubtype = srole[3]
    if rtype and ims == 'ims' and lis == 'lis':
        return rvocab, rtype, rsubtype
    else:
        raise ValueError("Badly formed lti role: %s" % str(role))


def is_subrole(role, parent_role):
    """True if role is a sub-role of parent_role

    role
        A :class:`~pyslet.urn.URN` instance containing the
        full definition of the role to be tested.

    parent_role
        A :class:`~pyslet.urn.URN` instance containing the full
        definition of the parent role.  It must *not* define a subrole
        of ValueError is raised.

    In the special case that role does not have subrole then it is
    simply matched against parent_role.  This ensures that::

        is_subrole(role, ROLE_HANDLES['Learner'])

    will return True in all cases where role is a Learner role."""
    rvocab, rtype, rsubtype = split_role(role)
    pvocab, ptype, psubtype = split_role(parent_role)
    if rvocab == pvocab and rtype == ptype and psubtype is None:
        return True
    return False


#: A mapping from LTI context type handles to the full URN for the
#: context type as a :class:`~pyslet.rfc2396.URI` instance.
CONTEXT_TYPE_HANDLES = {
    'CourseTemplate':
        URI.from_octets('urn:lti:context-type:ims/lis/CourseTemplate'),
    'CourseOffering':
        URI.from_octets('urn:lti:context-type:ims/lis/CourseOffering'),
    'CourseSection':
        URI.from_octets('urn:lti:context-type:ims/lis/CourseSection'),
    'Group': URI.from_octets('urn:lti:context-type:ims/lis/Group')
}


def load_metadata():
    """Loads the default metadata document

    Returns a :class:`pyslet.odata2.metadata.Document` instance. The
    schema is loaded from a bundled metadata document which contains the
    minimum schema required for an LTI tool provider."""
    mdir = os.path.split(os.path.abspath(__file__))[0]
    metadata = edmx.Document()
    with open(os.path.join(mdir, 'imsbltiv1p0_metadata.xml'), 'rb') as f:
        metadata.read(f)
    return metadata


class ToolConsumer(object):

    """An LTI consumer object

    entity
        An :class:`~pyslet.odata2.csdl.Entity` instance.

    cipher
        An :class:`~pyslet.wsgi.AppCipher` instance.

    This class is a light wrapper for the entity object that is used to
    persist information on the server about the consumer.  The consumer
    is persisted in a data store using a single entity passed on
    construction which must have the following required properties:

    ID: Int64
        A database key for the consumer.

    Handle: String
        A convenient handle for referring to this consumer in the user
        interface of the silo's owner.  This handle is never exposed to
        users launching the tool through the LTI protocol.  For example,
        you might use handles like "LMS Prod" and "LMS Staging" as handles
        to help distinguish different consumers.

    Key: String
        The consumer key

    Secret: String
        The consumer secret (encrypted using *cipher*).

    Silo: Entity
        Required navigation property to the Silo this consumer is
        associated with.

    Contexts: Entity Collection
        Navigation property to the associated contexts from which this
        tool has been launched.

    Resources: Entity Collection
        Navigation property to the associated resources from which this
        tool has been launched.

    Users: Entity Collection
        Navigation property to the associated users that have launched
        the tool."""

    def __init__(self, entity, cipher):
        #: the entity that persists this consumer
        self.entity = entity
        #: the cipher used to
        self.cipher = cipher
        #: the consumer key
        self.key = entity['Key'].value
        #: the consumer secret
        self.secret = self.cipher.decrypt(
            self.entity['Secret'].value).decode('utf-8')

    @classmethod
    def new_from_values(cls, entity, cipher, handle, key=None, secret=None):
        """Create an instance from an new entity

        entity
            An :class:`~pyslet.odata2.csdl.Entity` instance from a
            suitable entity set.

        cipher
            An :class:`~pyslet.wsgi.AppCipher` instance, used to encrypt
            the secret before storing it.

        handle
            A string

        key (optional)
            A text string, defaults to a string generated with
            :func:`~pyslet.wsgi.generate_key`

        secret (optional)
            A text string, defaults to a string generated with
            :func:`~pyslet.wsgi.generate_key`

        The fields of the entity are set from the passed in parameters
        (or the defaults) and then a new instance of *cls* is
        constructed from the entity and cipher and returned as a the
        result."""
        if secret is None:
            secret = wsgi.generate_key()
        if key is None:
            key = wsgi.generate_key()
        entity['ID'].set_from_value(wsgi.key60(key.encode('utf-8')))
        entity['Handle'].set_from_value(handle)
        entity['Key'].set_from_value(key)
        entity['Secret'].set_from_value(cipher.encrypt(secret.encode('utf-8')))
        return cls(entity, cipher)

    def update_from_values(self, handle, secret):
        """Updates an instance from new values

        handle
            A string used to update the consumer's handle

        secret
            A string used to update the consumer's secret

        It is not possible to update the consumer key as this is used to
        set the ID of the consumer itself."""
        self.entity['Handle'].set_from_value(handle)
        self.entity['Secret'].set_from_value(
            self.cipher.encrypt(secret.encode('utf-8')))
        self.entity.commit()
        # the base class only defines a constructor, so we have to hope
        # that it doesn't cache information about consumer secrets and
        # just update the value directly.
        self.secret = secret

    def nonce_key(self, nonce):
        """Returns a key into the nonce table

        nonce
            A string received as a nonce during an LTI launch.

        This method hashes the nonce, along with the consumer entity's
        *Key*, to return a hex digest string that can be used as a key
        for comparing against the nonces used in previous launches.

        Mixing the consumer entity's *Key* into the hash reduces the
        chance of a collision between two nonces from separate
        consumers."""
        return sha256(
            (self.entity['Key'].value + nonce).encode('utf-8')).hexdigest()

    def get_context(self, context_id, title=None, label=None, ctypes=None):
        """Returns a context entity

        context_id
            The context_id string passed on launch

        title (optional)
            The title string passed on launch

        label (optional)
            The label string passed on launch

        ctypes (optional)
            An array of :class:`~pyslet.rfc2396.URI` instances
            representing the context types of this context.  See
            :data:`CONTEXT_TYPE_HANDLES` for more information.

        Returns the context entity.

        If this context has never been seen before then a new entity is
        created and bound to the consumer.  Otherwise, the additional
        information (if supplied) is compared and updated as
        necessary."""
        with self.entity['Contexts'].open() as collection:
            key = wsgi.key60(
                (self.entity['Key'].value + context_id).encode('utf-8'))
            try:
                context = collection[key]
                update = False
                if title and context['Title'].value != title:
                    context['Title'].set_from_value(title)
                    update = True
                if label and context['Label'].value != label:
                    context['Label'].set_from_value(label)
                    update = True
                if ctypes:
                    ctypes.sort()
                    ctypes = ' '.join([str(x) for x in ctypes])
                    if context['Types'].value != ctypes:
                        context['Types'].set_from_value(ctypes)
                        update = True
                if update:
                    collection.update_entity(context)
            except KeyError:
                # first time we ever saw this context, create an entity
                context = collection.new_entity()
                context['ID'].set_from_value(key)
                context['ContextID'].set_from_value(context_id)
                context['Title'].set_from_value(title)
                context['Label'].set_from_value(label)
                if ctypes:
                    ctypes.sort()
                    ctypes = ' '.join([str(x) for x in ctypes])
                    context['Types'].set_from_value(ctypes)
                collection.insert_entity(context)
        return context

    def get_resource(self, resource_link_id, title=None, description=None,
                     context=None):
        """Returns a resource entity

        resource_link_id
            The resource_link_id string passed on launch (required).

        title (optional)
            The title string passed on launch, or None.

        description (optional)
            The description string passed on launch, or None.

        context (optional)
            The context entity referred to in the launch, or None.

        If this resource has never been seen before then a new entity is
        created and bound to the consumer and (if specified) the
        context.  Otherwise, the additional information (if supplied) is
        compared and updated as necessary, with the proviso that a
        resource can never change context, as per the following quote
        from the specification:

            [resource_link_id] will also change if the item is exported
            from one system or context and imported into another system
            or context. """
        with self.entity['Resources'].open() as collection:
            link = edm.EDMValue.from_type(edm.SimpleType.String)
            link.set_from_value(resource_link_id)
            filter = odata.CommonExpression.from_str("LinkID eq :link",
                                                     {'link': link})
            collection.set_filter(filter)
            resources = collection.values()
            if len(resources) == 0:
                # first time we ever saw this resource, create an entry
                resource = collection.new_entity()
                resource['LinkID'].set_from_value(resource_link_id)
                resource['Title'].set_from_value(title)
                resource['Description'].set_from_value(description)
                if context is not None:
                    # link this resource to this context
                    resource['Context'].bind_entity(context)
                collection.insert_entity(resource)
            elif len(resources) == 1:
                resource = resources[0]
                update = False
                if title and resource['Title'].value != title:
                    resource['Title'].set_from_value(title)
                    update = True
                if (description and resource['Description'].value !=
                        description):
                    resource['Description'].set_from_value(description)
                    update = True
                if update:
                    collection.update_entity(resource)
            else:
                logging.warn("Duplicate resource: %s:%s",
                             self.entity['Key'].value, resource_link_id)
                resource = resource[0]
        return resource

    def get_user(self, user_id, name_given=None, name_family=None,
                 name_full=None, email=None):
        """Returns a user entity

        user_id
            The user_id string passed on launch

        name_given
            The user's given name (or None)

        name_family
            The user's family name (or None)

        name_full
            The user's full name (or None)

        email
            The user's email (or None)

        If this user has never been seen before then a new entity is
        created and bound to the consumer, otherwise the """
        with self.entity['Users'].open() as collection:
            id = edm.EDMValue.from_type(edm.SimpleType.String)
            id.set_from_value(user_id)
            filter = odata.CommonExpression.from_str("UserID eq :id",
                                                     {'id': id})
            collection.set_filter(filter)
            users = collection.values()
            if len(users) == 0:
                # first time we ever saw this user, create an entry
                user = collection.new_entity()
                user['UserID'].set_from_value(user_id)
                user['GivenName'].set_from_value(name_given)
                user['FamilyName'].set_from_value(name_family)
                user['FullName'].set_from_value(name_full)
                user['Email'].set_from_value(email)
                collection.insert_entity(user)
            elif len(users) == 1:
                user = users[0]
                update = False
                if name_given and name_given != user['GivenName'].value:
                    user['GivenName'].set_from_value(name_given)
                    update = True
                if name_family and name_family != user['FamilyName'].value:
                    user['FamilyName'].set_from_value(name_family)
                    update = True
                if name_full and name_full != user['FullName'].value:
                    user['FullName'].set_from_value(name_full)
                    update = True
                if email and email != user['Email'].value:
                    user['Email'].set_from_value(email)
                    update = True
                if update:
                    collection.update_entity(user)
            else:
                logging.warn("Duplicate user: %s:%s",
                             self.entity['Key'].value, user_id)
                user = users[0]
        return user


class ToolProvider(oauth.RequestValidator, MigratedClass):

    """An LTI tool provider object

    consumers
        The :class:`~pyslet.odata2.csdl.EntitySet` containing the
        tool *Consumers*.

    nonces
        The :class:`~pyslet.odata2.csdl.EntitySet` containing the
        used *Nonces*.

    cipher
        An :class:`~pyslet.wsgi.AppCipher` instance.  Used to decrypt
        the consumer secret from the database.

    Implements the RequestValidator object required by the oauthlib
    package. Internally creates an instance of SignatureOnlyEndpoint"""

    def __init__(self, consumers, nonces, cipher):
        #: The entity set containing Silos
        self.consumers = consumers
        #: The entity set containing Nonces
        self.nonces = nonces
        #: The cipher object used for encrypting consumer secrets
        self.cipher = cipher
        self.endpoint = oauth.SignatureOnlyEndpoint(self)

    enforce_ssl = False

    def check_client_key(self, key):
        # any non-empty string is OK as a client key
        return len(key) > 0

    def check_nonce(self, nonce):
        # any non-empty string is OK as a nonce
        return len(nonce) > 0

    def validate_client_key(self, client_key, request):
        try:
            self.lookup_consumer(client_key)
            return True
        except KeyError:
            return False

    def validate_timestamp_and_nonce(self, client_key, timestamp, nonce,
                                     request, request_token=None,
                                     access_token=None):
        key = sha256((client_key + nonce).encode('utf-8')).hexdigest()
        with self.nonces.open() as collection:
            now = time.time()
            try:
                e = collection[key]
                last_seen = e['LastSeen'].value.with_zone(0).get_unixtime()
                if last_seen + 5400.0 < now:
                    # last seen more than 90 mins ago, update last_seen
                    e['LastSeen'].set_from_value(now)
                    collection.update_entity(e)
                    return True
                else:
                    # this is an error
                    return False
            except KeyError:
                e = collection.new_entity()
                e.set_key(key)
                e['LastSeen'].set_from_value(now)
                collection.insert_entity(e)
                return True

    dummy_client = ul(
        'dummy_'
        '6c877d7e0a8d52d3ea51155c0ce5bd75ceaf7bdd1d2041f9fb3703a207278ab9')

    dummy_secret = ul('secret')

    def get_client_secret(self, client_key, request):
        try:
            return self.lookup_consumer(client_key).secret
        except KeyError:
            # return the same value as the secret
            return self.dummy_secret

    def lookup_consumer(self, key):
        """Implements the required method for consumer lookup

        Returns a :class:`ToolConsumer` instance or raises a KeyError if
        key is not the key of any known consumer."""
        with self.consumers.open() as collection:
            return ToolConsumer(collection[wsgi.key60(force_bytes(key))],
                                self.cipher)

    @old_method('Launch')
    def launch(self, command, url, headers, body_string):
        """Checks a launch request for authorization

        command
            The HTTP method, as an upper-case string.  Should be POST
            for LTI.

        url
            The full URL of the page requested as part of the launch.
            This will be the launch URL specified in the LTI protocol
            and configured in the consumer.

        headers
            A dictionary of headers, must include the Authorization
            header but other values are ignored.

        body_string
            The query string (in the LTI case, this is the content of
            the POST request).

        Returns a :class:`ToolConsumer` instance and a dictionary of
        parameters on success. If the incoming request is not authorized
        then :class:`LTIAuthenticationError` is raised.

        This method also checks the LTI message type and protocol
        version and will raise :class:`LTIProtcolError` if this is not a
        recognized launch request."""
        try:
            result, request = self.endpoint.validate_request(
                url, command, body_string, headers)
            if not result:
                raise LTIAuthenticationError
            # grab the parameters as a dictionary from body_string
            parameters = parse_qs(body_string)
            for n, v in dict_items(parameters):
                parameters[n] = ','.join(v)
            # the consumer key is in the oauth_consumer_key param
            consumer = self.lookup_consumer(parameters['oauth_consumer_key'])
            message_type = parameters.get('lti_message_type', '')
            if (message_type.lower() != LTI_MESSAGE_TYPE):
                logging.warn("Unknown lti_message_type: %s", message_type)
                raise LTIProtocolError
            message_version = parameters.get('lti_version', '')
            if (message_version != LTI_VERSION):
                logging.warn("Unknown lti_version: %s", message_version)
                raise LTIProtocolError
            return consumer, parameters
        finally:
            pass


class ToolProviderContext(wsgi.SessionContext):

    def __init__(self, environ, start_response, canonical_root=None):
        wsgi.SessionContext.__init__(self, environ, start_response,
                                     canonical_root)
        #: a :class:`~pyslet.imsbltiv1p0.ToolConsumer` instance
        #: identified from the launch
        self.consumer = None
        #: a dictionary of non-oauth parameters from the launch
        self.parameters = {}
        #: the effective visit entity
        self.visit = None
        #: the effective resource entity
        self.resource = None
        #: the effective user entity
        self.user = None
        #: the effective group (context) entity
        self.group = None
        #: the effective permissions (an integer for bitwise testing)
        self.permissions = 0


class ToolProviderApp(wsgi.SessionApp):

    """Represents WSGI applications that provide LTI Tools

    The key 'ToolProviderApp' is reserved for settings defined by this
    class in the settings file. The defined settings are:

    silo ('testing')
        The name of a default silo to create when the --create_silo
        option is used.

    key ('12345')
        The default consumer key created when --create_silo is used.

    secret ('secret')
        The consumer secret of the default consumer created when
        --create_silo is used."""

    #: We have our own context class
    ContextClass = ToolProviderContext

    @classmethod
    def add_options(cls, parser):
        """Adds the following options:

        --create_silo       create default silo and consumer"""
        super(ToolProviderApp, cls).add_options(parser)
        parser.add_option(
            "--create_silo", dest="create_silo", action="store_true",
            default=False, help="Create default silo and consumer")

    @classmethod
    def setup(cls, options=None, args=None, **kwargs):
        super(ToolProviderApp, cls).setup(options, args, **kwargs)
        tp_settings = cls.settings.setdefault('ToolProviderApp', {})
        silo_name = tp_settings.setdefault('silo', 'testing')
        key = tp_settings.setdefault('key', '12345')
        secret = tp_settings.setdefault('secret', 'secret')
        if options and options.create_silo:
            # we need to create the default silo
            with cls.container['Silos'].open() as collection:
                silo = collection.new_entity()
                silo['Slug'].set_from_value(silo_name)
                collection.insert_entity(silo)
            cipher = cls.new_app_cipher()
            with silo['Consumers'].open() as collection:
                consumer = ToolConsumer.new_from_values(
                    collection.new_entity(), cipher, 'default', key=key,
                    secret=secret)
                collection.insert_entity(consumer.entity)

    @classmethod
    def load_default_metadata(cls):
        mdir = os.path.split(os.path.abspath(__file__))[0]
        metadata_file = os.path.abspath(
            os.path.join(mdir, 'imsbltiv1p0_metadata.xml'))
        metadata = edmx.Document()
        with open(metadata_file, 'rb') as f:
            metadata.read(f)
        return metadata

    def __init__(self, **kwargs):
        super(ToolProviderApp, self).__init__()
        self.provider = ToolProvider(
            self.container['Consumers'],
            self.container['Nonces'],
            self.app_cipher)
        self.stop = False

    def init_dispatcher(self):
        """Provides ToolProviderApp specific bindings.

        This method adds bindings for /launch as the launch URL for the
        tool and all paths within /resource as the resource pages
        themselves."""
        wsgi.SessionApp.init_dispatcher(self)
        self.set_method('/launch', self.lti_launch)
        self.set_method('/resource/*', self.resource_page)

    def set_launch_group(self, context):
        """Sets the group in the context from the launch parameters"""
        group_id = context.parameters.get('context_id', '')
        if not group_id:
            # optional parameter, but recommended
            return None
        group_types = context.parameters.get('context_type', '').split(',')
        gtypes = []
        for group_type in group_types:
            stype = group_type.strip()
            gtype_uri = URI.from_octets(stype)
            if not gtype_uri.is_absolute():
                gtype_uri = CONTEXT_TYPE_HANDLES.get(stype, None)
            gtypes.append(gtype_uri)
        group_title = context.parameters.get('context_title', '')
        group_label = context.parameters.get('context_label', '')
        context.group = context.consumer.get_context(
            group_id, group_title, group_label, gtypes)

    def set_launch_resource(self, context):
        """Sets the resource in the context from the launch parameters"""
        resource_id = context.parameters.get('resource_link_id', '')
        if not resource_id:
            # required parameter
            logging.warn("Missing resource_link_id")
            raise LTIProtocolError
        resource_title = context.parameters.get('resource_link_title', '')
        resource_description = context.parameters.get(
            'resource_link_description', '')
        # the unique resource within this consumer
        context.resource = context.consumer.get_resource(
            resource_id, resource_title, resource_description, context.group)

    def set_launch_user(self, context):
        """Sets the user in the context from the launch parameters"""
        user_id = context.parameters.get('user_id', '')
        if user_id:
            # not required
            context.user = context.consumer.get_user(
                user_id,
                name_given=context.parameters.get(
                    'lis_person_name_given', None),
                name_family=context.parameters.get(
                    'lis_person_name_family', None),
                name_full=context.parameters.get(
                    'lis_person_name_full', None),
                email=context.parameters.get(
                    'lis_person_contact_email_primary', None))
        else:
            context.user = None

    def set_launch_permissions(self, context):
        """Sets the permissions in the context from the launch params"""
        permissions = long2(0)
        if 'roles' in context.parameters:
            roles = context.parameters['roles'].split(',')
            for role in roles:
                srole = role.strip()
                role_uri = URI.from_octets(srole)
                if not role_uri.is_absolute():
                    role_uri = ROLE_HANDLES.get(srole, None)
                permissions |= self.get_permissions(role_uri)
        context.permissions = permissions

    #: Permission bit mask representing 'read' permission
    READ_PERMISSION = 0x1

    #: Permission bit mask representing 'write' permission
    WRITE_PERMISSION = 0x2

    #: Permission bit mask representing 'configure' permission
    CONFIGURE_PERMISSION = 0x4

    @classmethod
    def get_permissions(cls, role):
        """Returns the permissions that apply to a single role

        role
            A single :class:`~pyslet.urn.URN` instance

        Specific LTI tools can override this method to provide more
        complex permission models.  Each permission type is represented
        by an integer bit mask, permissions can be combined with binary
        or '|' to make an overal permissions integer.  The default
        implementation uses the :attr:`READ_PERMISSION`,
        :attr:`WRITE_PERMISSION` and :attr:`CONFIGURE_PERMISSION` bit
        masks but you are free to use any values you wish.

        In this implementation, Instructors (and all sub-roles) are
        granted read, write and configure whereas Learners (and all
        subroles) are granted read only.  Any other role returns 0
        (no permissions).

        An LTI consumer can specify multiple roles on launch, this
        method is called for *each* role and the resulting permissions
        integers are combined to provide an overall permissions
        integer."""
        logging.debug("Launch role: %s", str(role))
        if is_subrole(role, ROLE_HANDLES['Instructor']):
            return (cls.READ_PERMISSION | cls.WRITE_PERMISSION |
                    cls.CONFIGURE_PERMISSION)
        elif is_subrole(role, ROLE_HANDLES['Learner']):
            return cls.READ_PERMISSION
        return 0

    def get_user_display_name(self, context, user=None):
        """Given a user entity, returns a display name

        If user is None then the user from the context is used
        instead."""
        if user is None:
            user = context.user
        if user is None:
            return ''
        if user['FullName']:
            return user['FullName'].value
        elif user['GivenName'] and user['FamilyName']:
            return "%s %s" % (user['GivenName'].value,
                              user['FamilyName'].value)
        elif user['GivenName']:
            return user['GivenName'].value
        elif user['Email']:
            return user['Email'].value
        else:
            return ''

    def get_resource_title(self, context):
        """Given a resource entity, returns a display title"""
        resource = context.resource
        if resource is None:
            return ''
        if resource['Title']:
            return resource['Title'].value
        else:
            return 'Resource: %s' % resource['LinkID'].value

    def new_visit(self, context):
        """Called during launch to create a new visit entity

        A new visit entity is created and bound to the resource entity
        referred to in the launch.  The visit entity stores the
        permissions and a link to the (optional) user entity.

        If a visit to the same resource is already associated with the
        session it is replaced.  This ensures that information about the
        resource, the user, roles and permissions always corresponds to
        the most recent launch.

        Any visits from the same consumer but with a different user are
        also removed.  This handles the case where a previous user of
        the browser session needs to be logged out of the tool."""
        with self.container['Visits'].open() as collection:
            collection.set_expand(
                {'Resource': {'Consumer': None}, 'User': None})
            sid = edm.EDMValue.from_type(edm.SimpleType.String)
            sid.set_from_value(context.session.sid)
            filter = odata.CommonExpression.from_str(
                "Session eq :sid", {'sid': sid})
            collection.set_filter(filter)
            visits = collection.values()
            # now compare these visits to the new one
            for old_visit in visits:
                # if the old visit is to the same resource, replace it
                old_resource = old_visit['Resource'].get_entity()
                if old_resource.key() == context.resource.key():
                    # drop this visit
                    old_visit['Session'].set_from_value(None)
                    old_visit.expand(None, {'Session': None})
                    collection.update_entity(old_visit)
                    continue
                # if the old visit is to the same consumer but with a
                # different user, replace it
                old_consumer = old_resource['Consumer'].get_entity()
                if old_consumer.key() == context.consumer.entity.key():
                    old_user = old_visit['User'].get_entity()
                    # one or the other may be None, will compare by key
                    if context.user != old_user:
                        # drop this visit too
                        old_visit['Session'].set_from_value(None)
                        old_visit.expand(None, {'Session': None})
                        collection.update_entity(old_visit)
            collection.set_expand(None)
            collection.set_filter(None)
            visit = collection.new_entity()
            visit['Permissions'].set_from_value(context.permissions)
            visit['Session'].set_from_value(context.session.sid)
            visit['WhenLaunched'].set_from_value(iso.TimePoint.from_now_utc())
            visit['UserAgent'].set_from_value(
                context.environ.get('HTTP_USER_AGENT', ''))
            visit['Resource'].bind_entity(context.resource)
            user_value = []
            if context.user is not None:
                user_value.append(context.user)
                visit['User'].bind_entity(context.user)
            collection.insert_entity(visit)
            visit['Resource'].set_expansion_values([context.resource])
            visit['User'].set_expansion_values(user_value)
        context.visit = visit

    def find_visit(self, context, resource_id):
        """Finds a visit that matches this resource_id"""
        with self.container['Visits'].open() as collection:
            sid = edm.EDMValue.from_type(edm.SimpleType.String)
            sid.set_from_value(context.session.sid)
            filter = odata.CommonExpression.from_str(
                "Session eq :sid", {'sid': sid})
            collection.set_filter(filter)
            # we want to load the Resource, Resource/Consumer and User
            collection.set_expand({'Resource': {'Consumer': None},
                                   'User': None})
            visits = collection.values()
            for visit in visits:
                resource = visit['Resource'].get_entity()
                if resource.key() == resource_id:
                    return visit
        return None

    def establish_session(self, context):
        """Overridden to update the Session ID in the visit"""
        with self.container['Visits'].open() as collection:
            sid = edm.EDMValue.from_type(edm.SimpleType.String)
            sid.set_from_value(context.session.sid)
            filter = odata.CommonExpression.from_str(
                "Session eq :sid", {'sid': sid})
            collection.set_filter(filter)
            visits = collection.values()
            context.session.establish()
            # we don't expect more than one matching visit here
            for visit in visits:
                visit['Session'].set_from_value(context.session.sid)
                # just update the Session field
                visit.expand(None, {'Session': None})
                collection.update_entity(visit)

    def merge_session(self, context, merge_session):
        """Overridden to update the Session ID in any associated visits"""
        with self.container['Visits'].open() as collection:
            sid = edm.EDMValue.from_type(edm.SimpleType.String)
            sid.set_from_value(merge_session.session.sid)
            filter = odata.CommonExpression.from_str(
                "Session eq :sid", {'sid': sid})
            collection.set_filter(filter)
            collection.set_expand(
                {'Resource': {'Consumer': None}, 'User': None})
            merge_visits = collection.values()
            # there should be only one visit in this session if there
            # are more then something strange is going on
            sid.set_from_value(context.session.sid)
            for merge_visit in merge_visits:
                merge_resource = merge_visit['Resource'].get_entity()
                merge_consumer = merge_resource['Consumer'].get_entity()
                merge_user = merge_visit['User'].get_entity()
                old_visits = collection.values()
                for old_visit in old_visits:
                    # if the old visit is to the same resource, replace it
                    old_resource = old_visit['Resource'].get_entity()
                    if old_resource.key() == merge_resource.key():
                        # drop this visit's session
                        old_visit['Session'].set_from_value(None)
                        old_visit.expand(None, {'Session': None})
                        collection.update_entity(old_visit)
                        continue
                    # if the old visit is to the same consumer but with a
                    # different user, replace it
                    old_consumer = old_resource['Consumer'].get_entity()
                    if old_consumer.key() == merge_consumer.key():
                        old_user = old_visit['User'].get_entity()
                        # one or the other may be None, will compare by key
                        if old_user != merge_user:
                            # drop this visit too
                            old_visit['Session'].set_from_value(None)
                            old_visit.expand(None, {'Session': None})
                            collection.update_entity(old_visit)
                merge_visit['Session'].set_from_value(context.session.sid)
                # just update the Session field
                merge_visit.set_expand(None, {'Session': None})
                collection.update_entity(merge_visit)

    def load_visit(self, context):
        """Loads an existing LTI visit into the context

        You'll normally call this method from each session decorated
        method of your tool provider that applies to a protected
        resource.

        This method sets the following attributes of the context...

        :attr:`ToolProviderContext.resource`
            The resource record is identified from the resource
            id given in the URL path.

        :attr:`ToolProviderContext.visit`
            The session is searched for a visit record matching the
            resource.

        :attr:`ToolProviderContext.permissions`
            Set from the visit record

        :attr:`ToolProviderContext.user`
            The optional user is loaded from the visit.

        :attr:`ToolProviderContext.group`
            The context record identified from the resource id given in
            the URL path.  This may be None if the resource link was not
            created in any context.

        :attr:`ToolProviderContext.consumer`
            The consumer object is looked up from the visit entity.

        If the visit can't be set then an exception is raised, an
        unknown resource raises :class:`pyslet.wsgi.PageNotFound`
        whereas the absence of a valid visit for a known resource raises
        :class:`pyslet.wsgi.PageNotAuthorized`.  These are caught
        automatically by the WSGI handlers and return 404 and 403 errors
        respectively."""
        if context.visit is None:
            path = context.environ['PATH_INFO'].split('/')
            if len(path) < 4:
                raise wsgi.PageNotFound
            resource_id = path[2]
            try:
                resource_id = int(resource_id, 16)
            except ValueError:
                raise wsgi.PageNotFound
            context.visit = self.find_visit(context, resource_id)
            if context.visit is None:
                raise wsgi.PageNotAuthorized
            context.permissions = context.visit['Permissions'].value
            context.resource = context.visit['Resource'].get_entity()
            context.user = context.visit['User'].get_entity()
            context.group = context.resource['Context'].get_entity()
            tc_entity = context.resource['Consumer'].get_entity()
            context.consumer = ToolConsumer(tc_entity, self.app_cipher)

    def lti_launch(self, context):
        # we are only interested in the authorisation header
        h = context.environ.get('CONTENT_TYPE', None)
        if h:
            headers = {'Content-Type': h}
        else:
            headers = None
        try:
            context.consumer, context.parameters = self.provider.launch(
                context.environ['REQUEST_METHOD'].upper(),
                str(context.get_url()), headers, context.get_content())
            self.set_launch_group(context)
            self.set_launch_resource(context)
            self.set_launch_user(context)
            self.set_launch_permissions(context)
            # load the session information into the context and add a
            # new visit to it.
            self.set_session(context)
            self.new_visit(context)
        except LTIProtocolError:
            logging.exception("LTI Protocol Error")
            return self.error_page(context, 400)
        except LTIAuthenticationError:
            logging.exception("LTI Authentication Failure")
            return self.error_page(context, 403)
        launch_target = URI.from_octets(
            "resource/%08X/" %
            context.resource.key()).resolve(context.get_app_root())
        return self.session_page(context, self.launch_redirect, launch_target)

    def launch_redirect(self, context):
        """Redirects to the resource identified on launch

        A POST request should pretty much always redirect to a GET page
        and our tool launches are no different.  This allows you to
        reload a tool page straight away if desired without the risk of
        double-POST issues."""
        resource_link = URI.from_octets(
            "resource/%08X/" %
            context.resource.key()).resolve(context.get_app_root())
        return self.redirect_page(context, resource_link, 303)

    @wsgi.session_decorator
    def resource_page(self, context):
        """Returns a resource page

        This method is the heart of your tool and will be called for all
        URLs that start with a resource path.  You can override the
        behaviour for specific paths by adding new methods to the
        dispatcher with more specific URLs, such as
        /resource/*/view.html, but anything in /resource/*/* that is
        not matched by a more specific method will end up here.

        You'll typically call :meth:`load_visit` to load information
        about the associated LTI visit into the context before proceeding,
        it checks authorisation to view the resource for you.

        The default implementation returns a simple page with the
        resource's title and a simple message reflecting the user name
        of the authorised user."""
        page = """<!DOCTYPE html PUBLIC
    "-//W3C//DTD XHTML 1.0 Strict//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html>
  <head>
   <title>%(title)s</title>
  </head>
<body>
<h4>%(title)s</h4>
<p>Congratulations %(user)s, you've launched an LTI tool created with <a
    href="http://www.pyslet.org/" target="_blank">Pyslet</a>.</p>
</body>
</html>"""
        self.load_visit(context)
        params = {'user':
                  xml.escape_char_data7(self.get_user_display_name(context)),
                  'title':
                  xml.escape_char_data7(self.get_resource_title(context))
                  }
        data = page % params
        context.set_status(200)
        return self.html_response(context, data)


# legacy definitions
BLTI_VERSION = LTI_VERSION
BLTI_LAUNCH_REQUEST = LTI_MESSAGE_TYPE

BLTIError = LTIError
BLTIOAuthParameterError = LTIAuthenticationError
BLTIAuthenticationError = LTIAuthenticationError


class BLTIDuplicateKeyError(BLTIError):
    pass


class BLTIConsumer(ToolConsumer):
    pass


class BLTIToolProvider(ToolProvider):

    """Legacy class for tool provider.

    Refactored to build directly on the newer :class:`ToolProvider`. A
    single Silo entity is created containing all defined consumers. An
    in-memory SQLite database is used as the data store.  Consumer keys
    are not encrypted (a plaintext cipher is used) as they will not be
    persisted."""

    def __init__(self):
        # initialise a very simple in-memory databse
        metadata = load_metadata()
        self.container = metadata.root.DataServices.defaultContainer
        self.data_source = sql.SQLiteEntityContainer(
            file_path=':memory:', container=self.container)
        self.data_source.create_all_tables()
        cipher = wsgi.AppCipher(0, 'secret', self.container['AppKeys'])
        with self.container['Silos'].open() as collection:
            self.silo = collection.new_entity()
            self.silo['ID'].set_from_value(wsgi.key60(b'BLTIToolProvider'))
            self.silo['Slug'].set_from_value('BLTIToolProvider')
            collection.insert_entity(self.silo)
        ToolProvider.__init__(self, self.container['Consumers'],
                              self.container['Nonces'], cipher)

    @old_method('GenerateKey')
    def generate_key(self, key_length=128):
        """Generates a new key

        Also available as GenerateKey.  This method is deprecated, it
        has been replaced by the similarly named function
        :func:`pyslet.wsgi.generate_key`.

        key_length
            The minimum key length in bits.  Defaults to 128.

        The key is returned as a sequence of 16 bit hexadecimal
        strings separated by '.' to make them easier to read and
        transcribe into other systems."""
        key = []
        nfours = (key_length + 1) // 16
        try:
            rbytes = os.urandom(nfours * 2)
            for i in range3(nfours):
                four = "%02X%02X" % (
                    byte_value(rbytes[2 * i]), byte_value(rbytes[2 * i + 1]))
                key.append(four)
        except NotImplementedError:
            for i in range3(nfours):
                four = []
                for j in range3(4):
                    four.append(random.choice('0123456789ABCDEFG'))
                key.append(''.join(four))
        return '.'.join(key)

    @old_method('NewConsumer')
    def new_consumer(self, key=None, secret=None):
        """Creates a new BLTIConsumer instance

        Also available as NewConsumer

        The new instance is added to the database of consumers
        authorized to use this tool.  The consumer key and secret
        are automatically generated using :meth:`generate_key` but
        key and secret can be passed as optional arguments instead."""
        if key is None:
            key = self.generate_key()
        if secret is None:
            secret = self.generate_key()
        try:
            with self.silo['Consumers'].open() as collection:
                consumer = ToolConsumer.new_from_values(
                    collection.new_entity(), self.cipher, key, key=key,
                    secret=secret)
                collection.insert_entity(consumer.entity)
        except edm.ConstraintError:
            raise BLTIDuplicateKeyError(key)
        return key, secret

    @old_method('LoadFromFile')
    def load_from_file(self, f):
        """Loads the list of trusted consumers

        Also available as LoadFromFile

        The consumers are loaded from a simple file of key, secret
        pairs formatted as::

            <consumer key> [SPACE]+ <consumer secret>

        Lines starting with a '#' are ignored as comments."""
        lines = f.readlines()
        for line in lines:
            if line and line[0] == '#':
                continue
            fields = line.split()
            if len(fields) >= 2:
                key = fields[0].decode('utf-8')
                secret = fields[1].decode('utf-8')
                try:
                    self.lookup_consumer(key)
                    raise BLTIDuplicateKeyError(key)
                except KeyError:
                    self.new_consumer(key, secret)

    @old_method('SaveToFile')
    def save_to_file(self, f):
        """Saves the list of trusted consumers

        Also available as SaveToFile

        The consumers are saved in a simple file suitable for
        reading with :meth:`load_from_file`."""
        with self.silo['Consumers'].open() as collection:
            consumers = {}
            for c in collection.itervalues():
                consumer = ToolConsumer(c, self.cipher)
                consumers[consumer.key] = consumer.secret
            keys = sorted(dict_keys(consumers))
            for key in keys:
                f.write(("%s %s\n" % (key, consumers[key])).encode('ascii'))
